from math import ceil, floor

import pytest
import torch

from elasticai.creator.nn.fixed_point.mac.layer import MacLayer
from elasticai.creator.nn.fixed_point.number_converter import FXPParams
from elasticai.creator.vhdl.ghdl_simulation import GHDLSimulator

integer_test_data = [
    (FXPParams(4, 0), x1, x2)
    for x1, x2 in [
        ((1.0, 0.0), (-1.0, 0.0)),
        ((1.0, 0.0), (0.0, 0.0)),
        ((1.0, 0.0), (3.0, 0.0)),
        ((2.0, 0.0), (4.0, 0.0)),
        ((-2.0, 0.0), (4.0, 0.0)),
        ((-8.0, -8.0), (7.0, 7.0)),
        ((1.0, 1.0), (1.0, 1.0)),
    ]
]

fractions_test_data = [
    (FXPParams(5, 2), x1, x2)
    for x1, x2 in [
        # 00010 * 00010 -> 00000 00100 -> 000(00 001)00 -> 00001
        ((0.5, 0.0), (0.5, 0.0)),
        # 00010 * 01000 -> 00000 10000 -> 000(00 100)00 -> 00100
        ((0.5, 0.0), (2.0, 0.0)),
        ((0.25, 0.0), (0.5, 0.0)),
        ((0.5, 0.5), (0.5, 0.5)),
        # 00001 * 00010 + 00100 * 00001 -> 00000 00010 + 00000 00100 -> 00000 0110 -> 00(000 01)10 -> 00(000 10)00 -> 00010
        ((0.25, 1.0), (0.5, 0.25)),
        ((-0.25, -1.0), (0.5, 0.5)),
        ((-4.0, -4.0), (3.75, 3.75)),
    ]
]


@pytest.mark.skip(
    reason="Not changed for new SimulatedLayer. Needs to be implemented in the future"
)
@pytest.mark.simulation
@pytest.mark.parametrize(
    ["fxp_params", "x1", "x2"], integer_test_data + fractions_test_data
)
def test_mac_hw_for_integers(tmp_path, fxp_params, x1, x2):
    root_dir_path = str(tmp_path)
    mac = MacLayer(fxp_params=fxp_params, vector_width=2)
    y = mac(torch.tensor(x1), torch.tensor(x2)).item()
    sim = mac.create_simulation(GHDLSimulator, root_dir_path)
    actual = sim(x1, x2)
    assert y == actual


def macop_in_hw_modelled_via_native_python(
    a_v: list[float], b_v: list[float], total_bits: int, frac_bits: int
) -> float:
    max_int = 2 ** (total_bits - 1 - frac_bits) - 1
    min_int = -1 * 2 ** (total_bits - 1 - frac_bits)

    def clamp(x):
        return min(max_int, max(min_int, x))

    def round(x):
        if x < 0:
            return ceil(x)
        else:
            return floor(x)

    def to_int(x):
        return int(x * 2**frac_bits)

    ys = [a * b for a, b in zip(map(to_int, a_v), map(to_int, b_v))]
    y = sum(ys)
    y = y / (2**frac_bits)
    y = clamp(y)
    y = round(y)
    return y / (2**frac_bits)


@pytest.mark.parametrize(
    "x1, x2",
    [
        # 00001 * 00010 + 00100 * 00001 -> 00000 00010 + 00000 00100 -> 00000 00110 -> 000(00 001)10 -> 00001
        ((0.25, 1.0), (0.5, 0.25)),
        # 11111 * 00010 + 11100 * 00010 -> 11111 11111 * 00000 00010 + 11111 11100 * 00000 00010
        #  ->  11111 11110
        #    + 11111 11000
        #   -> 11111 10110 -> 111(11 101)10 -> rounding up: 111(11 110)10 -> 11110
        ((-0.25, -1.0), (0.5, 0.5)),
    ],
)
def test_sw_mac_rounds_towards_zero(x1, x2):
    fxp_params = FXPParams(total_bits=5, frac_bits=2)
    expected = macop_in_hw_modelled_via_native_python(x1, x2, frac_bits=2, total_bits=5)
    mac = MacLayer(fxp_params=fxp_params, vector_width=len(x1))
    y = mac(torch.tensor(x1), torch.tensor(x2)).item()
    assert expected == y



from elasticai.creator.ir.base import Attribute, IrData, ReadOnlyField


def _read_only_str() -> ReadOnlyField[str, str]:
    return ReadOnlyField(lambda x: x)


class Node(IrData):
    """
    NOTE: name and type are read only when accessed through their descriptor, i.e.
    >>> n = Node(dict())
    >>> n.name = "x" #  Error! read only
    >>> x = n.name  #  Error! no key `name` in n.data
    >>> n.data['name'] = "x"
    >>> n.name #  Ok!
    ... 'x'
    """

    name: ReadOnlyField[str, str] = _read_only_str()
    type: ReadOnlyField[str, str] = _read_only_str()


class Edge(IrData):
    """
    NOTE: src, sink are read only when accessed through their descriptors.
    """

    src: ReadOnlyField[str, str] = _read_only_str()
    sink: ReadOnlyField[str, str] = _read_only_str()


def node(name: str, type: str, attributes: dict[str, Attribute] | None = None) -> Node:
    if attributes is None:
        attributes = dict()
    return Node(dict(name=name, type=type, **attributes))


def edge(src: str, sink: str, attributes: dict[str, Attribute] | None = None) -> Edge:
    if attributes is None:
        attributes = dict()
    return Edge(dict(src=src, sink=sink, **attributes))



from elasticai.creator.base_modules.hard_sigmoid import HardSigmoid as HardSigmoidBase
from elasticai.creator.nn.design_creator_module import DesignCreatorModule
from elasticai.creator.nn.fixed_point.two_complement_fixed_point_config import (
    FixedPointConfig,
)

from .design import HardSigmoid as HardSigmoidDesign


class HardSigmoid(DesignCreatorModule, HardSigmoidBase):
    def __init__(self, total_bits: int, frac_bits: int) -> None:
        super().__init__()
        self._config = FixedPointConfig(total_bits=total_bits, frac_bits=frac_bits)

    def create_design(self, name: str) -> HardSigmoidDesign:
        return HardSigmoidDesign(
            name=name,
            total_bits=self._config.total_bits,
            frac_bits=self._config.frac_bits,
            one=self._config.as_integer(1),
            zero_threshold=self._config.as_integer(-3),
            one_threshold=self._config.as_integer(3),
            slope=self._config.as_integer(1 / 6),
            y_intercept=self._config.as_integer(0.5),
        )



import re


class NameRegistry:
    def __init__(self):
        self._registry = {}

    def prepopulate(self, names):
        for name in names:
            match = re.match(r"(.+)_(\d+)$", name)
            if match:
                name = match.group(1)
                suffix = int(match.group(2))
            else:
                suffix = 0
            self._registry[name] = self._registry.get(name, suffix) + 1
        return self

    def get_unique_name(self, name):
        if name not in self._registry:
            self._registry[name] = 0
            return name

        new_name = f"{name}_{self._registry[name]}"
        self._registry[name] += 1
        return new_name



from elasticai.creator.file_generation.savable import Path
from elasticai.creator.file_generation.template import (
    InProjectTemplate,
    module_to_package,
)
from elasticai.creator.vhdl.code_generation.addressable import calculate_address_width
from elasticai.creator.vhdl.code_generation.code_abstractions import (
    to_vhdl_binary_string,
)


class Rom:
    def __init__(
        self, name: str, data_width: int, values_as_integers: list[int]
    ) -> None:
        self._name = name
        self._data_width = data_width
        number_of_values = len(values_as_integers)
        self._address_width = self._bits_required_to_address_n_values(number_of_values)
        self._values = [
            to_vhdl_binary_string(x, self._data_width)
            for x in self._append_zeros_to_fill_addressable_memory(values_as_integers)
        ]

    def save_to(self, destination: Path):
        template = InProjectTemplate(
            file_name="rom.tpl.vhd",
            package=module_to_package(self.__module__),
            parameters=dict(
                rom_value=self._rom_values(),
                rom_addr_bitwidth=str(self._address_width),
                rom_data_bitwidth=str(self._data_width),
                name=self._name,
                resource_option="auto",
            ),
        )
        destination.as_file(".vhd").write(template)

    def _rom_values(self) -> str:
        return ",".join(self._values)

    def _append_zeros_to_fill_addressable_memory(self, values: list[int]) -> list[int]:
        missing_number_of_zeros = 2**self._address_width - len(values)
        return values + _zeros(missing_number_of_zeros)

    def _bits_required_to_address_n_values(self, n: int) -> int:
        return calculate_address_width(n)


def _zeros(n: int) -> list[int]:
    return [0] * n



from abc import abstractmethod
from collections import defaultdict
from typing import Protocol

from elasticai.creator.file_generation.savable import Path
from elasticai.creator.file_generation.template import (
    InProjectTemplate,
    module_to_package,
)
from elasticai.creator.nn.fixed_point.number_converter import FXPParams, NumberConverter
from elasticai.creator.vhdl.design.ports import Port
from elasticai.creator.vhdl.simulated_layer import Testbench


class LinearDesignProtocol(Protocol):
    @property
    @abstractmethod
    def name(self) -> str: ...

    @property
    @abstractmethod
    def port(self) -> Port: ...

    @property
    @abstractmethod
    def in_feature_num(self) -> int: ...

    @property
    @abstractmethod
    def out_feature_num(self) -> int: ...

    @property
    @abstractmethod
    def frac_width(self) -> int: ...

    @property
    @abstractmethod
    def data_width(self) -> int: ...


class LinearTestbench(Testbench):
    def __init__(self, name: str, uut: LinearDesignProtocol):
        self._converter_for_batch = NumberConverter(
            FXPParams(8, 0)
        )  # max for 255 lines of inputs
        self._name = name
        self._uut_name = uut.name
        self._input_signal_length = uut.in_feature_num
        self._x_address_width = uut.port["x_address"].width
        self._fxp_params = FXPParams(uut.data_width, uut.frac_width)
        self._converter = NumberConverter(self._fxp_params)
        self._output_signal_length = uut.out_feature_num
        self._y_address_width = uut.port["y_address"].width

    def save_to(self, destination: Path):
        template = InProjectTemplate(
            package=module_to_package(self.__module__),
            file_name="testbench.tpl.vhd",
            parameters={
                "testbench_name": self.name,
                "input_signal_length": str(self._input_signal_length),
                "total_bits": str(self._fxp_params.total_bits),
                "x_address_width": str(self._x_address_width),
                "output_signal_length": str(self._output_signal_length),
                "y_address_width": str(self._y_address_width),
                "uut_name": self._uut_name,
            },
        )
        destination.create_subpath(self.name).as_file(".vhd").write(template)

    @property
    def name(self) -> str:
        return self._name

    def prepare_inputs(self, *inputs) -> list[dict]:
        batches = inputs[0]
        prepared_inputs = []
        for batch in batches:
            prepared_inputs.append({})
            for channel_id, channel in enumerate(batch):
                for time_step_id, time_step_val in enumerate(channel):
                    prepared_inputs[-1][f"x_{channel_id}_{time_step_id}"] = (
                        self._converter.rational_to_bits(time_step_val)
                    )
        return prepared_inputs

    def parse_reported_content(self, content: list[str]) -> list[list[list[float]]]:
        """
        This function parses the reported content, which is just a list of strings.
        All lines starting with 'output_text:' are considered as a result of the testbench.
        These results will be stacked for each batch.
        So you get a list[list[list[float]]] which is similar to batch[out channels[output neurons[float]]].
        For linear layer the output neurons is 1.
        For each item reported it is checked if the string starts with 'result: '.
        If so the remaining part will be split by ','. The first number gives the batch. The second the result.
        """

        def split_list(a_list):
            print("len(a_list): ", len(a_list))
            new_list = list()
            new_list.append(list())
            for i, value in enumerate(a_list):
                new_list[0].append(value)
            return new_list

        results_dict = defaultdict(list)

        print()
        for line in map(str.strip, content):
            if line.startswith("result: "):
                batch_text = line.split(":")[1].split(",")[0][1:]
                output_text = line.split(":")[1].split(",")[1][0:]
                print("output_text: ", output_text)
                batch = int(self._converter_for_batch.bits_to_rational(batch_text))
                if "U" not in line.split(":")[1].split(",")[1][1:]:
                    output = self._converter.bits_to_rational(output_text)
                else:
                    output = output_text
                results_dict[batch].append(output)
            else:
                print(line)
        results = list()
        for x in results_dict.items():
            results.append(split_list(x[1]))
        print("results: ", results)
        if len(results) == 0:
            raise Exception(content)
        return list(results)



from elasticai.creator.ir2verilog import TemplateDirector


class TestVerilogTemplate:
    def test_can_set_localparam(self):
        template = (
            TemplateDirector()
            .set_prototype("localparam DATA_WIDTH = 4'd8;")
            .localparam("DATA_WIDTH")
            .build()
        )

        assert (
            template.substitute({"DATA_WIDTH": "'h16"})
            == "localparam DATA_WIDTH = 'h16;"
        )

    def test_can_set_parameter_before_comma(self):
        template = (
            TemplateDirector()
            .set_prototype("parameter DATA_WIDTH = 4'd8,")
            .parameter("DATA_WIDTH")
            .build()
        )

        assert (
            template.substitute({"DATA_WIDTH": "'h16"})
            == "parameter DATA_WIDTH = 'h16,"
        )

    def test_keep_new_line_between_define_switches(self):
        template = (
            TemplateDirector()
            .set_prototype("`define DATA_WIDTH\n`define DATA_LENGTH")
            .define_scoped_switch("DATA_WIDTH")
            .define_scoped_switch("DATA_LENGTH")
            .build()
        )

        assert (
            template.substitute(
                {
                    "DATA_WIDTH": True,
                    "DATA_LENGTH": True,
                }
            )
            == "`define DATA_WIDTH\n`define DATA_LENGTH"
        )

    def test_can_set_parameter_before_new_line(self):
        template = (
            TemplateDirector()
            .set_prototype("""parameter DATA_WIDTH = 4'd8
""")
            .parameter("DATA_WIDTH")
            .build()
        )

        assert (
            template.substitute({"DATA_WIDTH": "'h16"})
            == "parameter DATA_WIDTH = 'h16\n"
        )

    def test_can_undefine_data_width(self):
        template = (
            TemplateDirector()
            .set_prototype("`define DATA_WIDTH 8")
            .define_scoped_switch("DATA_WIDTH")
            .build()
        )
        template.undef("DATA_WIDTH")
        assert (
            template.substitute({})
            == """// automatically disabled\n// `define DATA_WIDTH 8"""
        )

    def test_can_leave_data_width_defined(self):
        template = (
            TemplateDirector()
            .set_prototype("`define DATA_WIDTH 8")
            .define_scoped_switch("DATA_WIDTH")
            .build()
        )
        template.define("DATA_WIDTH")
        assert template.substitute({}) == "`define DATA_WIDTH 8"

    def test_can_replace_module_name(self):
        template = (
            TemplateDirector()
            .set_prototype("module FILTER_FIR_HALF#(")
            .add_module_name()
            .build()
        )
        assert (
            template.substitute({"module_name": "FILTER_FIR_MY_NAME_HALF"})
            == "module FILTER_FIR_MY_NAME_HALF#("
        )

    def test_automatically_use_module_prefix_for_defines(self):
        template = (
            TemplateDirector()
            .set_prototype("`define DATA_WIDTH 8")
            .define_scoped_switch("DATA_WIDTH")
            .build()
        )
        assert (
            template.substitute({"module_name": "FILTER_FIR_HALF"})
            == "`define FILTER_FIR_HALF_DATA_WIDTH 8"
        )

    def test_later_occurences_of_defined_names_are_scoped(self):
        template = (
            TemplateDirector()
            .set_prototype("`define DATA_WIDTH 8\nDATA_WIDTH + 10")
            .define_scoped_switch("DATA_WIDTH")
            .build()
        )
        assert (
            template.substitute({"module_name": "FILTER_FIR_HALF"})
            == "`define FILTER_FIR_HALF_DATA_WIDTH 8\nFILTER_FIR_HALF_DATA_WIDTH + 10"
        )



from abc import abstractmethod
from collections.abc import Iterator, Mapping, Set
from typing import Protocol, Self, TypeVar

T = TypeVar("T", str, int)


class Graph(Protocol[T]):
    @property
    @abstractmethod
    def nodes(self) -> Set[T]: ...

    @abstractmethod
    def iter_edges(self) -> Iterator[tuple[T, T]]: ...

    @abstractmethod
    def add_node(self, node: T) -> Self: ...

    @abstractmethod
    def add_edge(self, src: T, sink: T) -> Self: ...

    @property
    @abstractmethod
    def predecessors(self) -> Mapping[T, set[T]]: ...

    @property
    @abstractmethod
    def successors(self) -> Mapping[T, set[T]]: ...

    @abstractmethod
    def new(self) -> "Graph[T]": ...



import pytest

from elasticai.creator.ir2vhdl import LoweringPass, Shape, edge, vhdl_node
from elasticai.creator_plugins.time_multiplexed_sequential.src import (
    network as _network_handler,
)
from elasticai.creator_plugins.time_multiplexed_sequential.src import (
    sequential,
)

from .high_level_network import network


@pytest.fixture
def lower():
    _lower = LoweringPass()
    _lower.register(sequential)
    _lower.register_iterable(_network_handler)
    return _lower


def test_sets_input_correctly_for_version_without_sliding_window(lower):
    lowered = tuple(
        lower(
            [
                network(
                    input_shape=(1, 3),
                    kernel_size=3,
                    out_channels=1,
                ),
            ]
        )
    )[0]
    expected_input_node = vhdl_node(
        name="input",
        type="input",
        implementation="",
        input_shape=(1, 3),
        output_shape=(1, 3),
    )
    assert lowered.nodes["input"] == expected_input_node


def test_first_shift_reg_has_three_output_wires():
    in_channels = 2
    length = 4
    _net = network(
        input_shape=(in_channels, length), kernel_size=2, out_channels=2, stride=1
    )

    lowered = sequential(_net)
    node = lowered.nodes["shift_register_i0"]
    assert node.output_shape == (2, 2)


def test_shift_reg_predecessor_is_conv1():
    in_channels = 2
    length = 4
    _net = network(
        input_shape=(in_channels, length), kernel_size=2, out_channels=2, stride=1
    )

    lowered = sequential(_net)
    p = {n.name for n in (iter(lowered.successors("shift_register_i0").values()))}

    assert p == {"conv1_i0"}


def test_second_conv_has_three_input_wires():
    kernel_size = 1
    _net = network(
        input_shape=(3, 2), kernel_size=kernel_size, out_channels=3, stride=1
    )
    hl_conv1 = _net.nodes["conv1_i0"]
    lowered = sequential(_net)
    low_level_conv1 = lowered.nodes["conv1_i0"]
    expected_shape = Shape(hl_conv1.input_shape.depth, kernel_size)
    assert low_level_conv1.input_shape == expected_shape


class TestsForSecondConv:
    def network(self):
        return network(input_shape=(3, 2), kernel_size=1, out_channels=3, stride=1)

    def get_hl_conv(self):
        pass

    def test_high_and_low_level_have_same_in_channels(self):
        _net = self.network()
        hl_conv1 = _net.nodes["conv1_i0"]
        lowered = sequential(_net)
        low_level_conv1 = lowered.nodes["conv1_i0"]
        assert low_level_conv1.input_shape.depth == hl_conv1.input_shape.depth


def test_add_correct_edges(lower):
    net = network(
        input_shape=(2, 7),
        kernel_size=3,
        out_channels=1,
    )
    lowered = tuple(lower([net]))[0]
    edges = tuple(lowered.edges.values())
    expected = (
        edge(
            src="input",
            sink="conv0_i0",
            src_sink_indices=tuple(),
        ),
        edge(
            src="conv0_i0",
            sink="striding_shift_register_i0",
            src_sink_indices=tuple(),
        ),
        edge(
            src="striding_shift_register_i0",
            sink="conv1_i0",
            src_sink_indices=tuple(),
        ),
        edge(src="conv1_i0", sink="output", src_sink_indices=tuple()),
    )
    assert edges == expected


def test_keeps_input_shape_for_first_conv(lower):
    lowered = tuple(
        lower(
            [
                network(
                    input_shape=(1, 4),
                    kernel_size=3,
                    out_channels=1,
                ),
            ]
        )
    )[0]
    input_shape = lowered.nodes["conv0_i0"].input_shape
    assert input_shape == lowered.nodes["input"].output_shape


def test_generates_adapter_for_addressing_skeleton(lower):
    lowered = {
        i.name: i
        for i in lower(
            [
                network(
                    input_shape=(1, 4),
                    kernel_size=4,
                    out_channels=1,
                ),
            ]
        )
    }

    adapter = lowered["buffered_network_wrapper"]
    assert adapter.attributes["generic_map"] == {
        "KERNEL_SIZE": "4",
        "STRIDE": "2",
    }


def test_generates_adapter_for_addressing_skeleton_with_varied_parameters(lower):
    in_channels = 2
    width = 10
    kernel_size = 3
    expected_adapter_kernel_size = kernel_size * in_channels
    expected_adapter_stride = in_channels
    lowered = {
        i.name: i
        for i in lower(
            [
                network(
                    input_shape=(in_channels, width),
                    kernel_size=kernel_size,
                    out_channels=4,
                    stride=1,
                ),
            ]
        )
    }

    adapter = lowered["buffered_network_wrapper"]
    assert adapter.attributes["generic_map"] == {
        "KERNEL_SIZE": str(expected_adapter_kernel_size),
        "STRIDE": str(expected_adapter_stride),
    }



from collections.abc import Callable, Iterable, Iterator, Mapping
from typing import (
    Any,
    Generic,
    ParamSpec,
    Self,
    TypeVar,
    cast,
    overload,
)

from elasticai.creator.graph import Graph
from elasticai.creator.ir.base import Attribute, IrData, read_only_field

from .core import Edge, Node

N = TypeVar("N", bound=Node, covariant=True)
E = TypeVar("E", bound=Edge, covariant=True)

StoredT = TypeVar("StoredT", bound=Attribute)
VisibleT = TypeVar("VisibleT")
P = ParamSpec("P")


class Implementation(IrData, Generic[N, E], create_init=False):
    __slots__ = (
        "data",
        "graph",
        "_node_data",
        "_edge_data",
        "_node_fn",
        "_edge_fn",
    )

    name: str
    type: str

    @overload
    def __init__(
        self: "Implementation[N, E]",
        *,
        graph: Graph[str],
        edge_fn: Callable[[dict[str, Attribute]], E],
        node_fn: Callable[[dict[str, Attribute]], N],
        data: dict[str, Attribute] | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self: "Implementation[N, Edge]",
        *,
        graph: Graph,
        node_fn: Callable[[dict[str, Attribute]], N],
        data: dict[str, Attribute] | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self: "Implementation[Node, E]",
        *,
        graph: Graph[str],
        edge_fn: Callable[[dict[str, Attribute]], E],
        data: dict[str, Attribute] | None = None,
    ) -> None: ...

    @overload
    def __init__(
        self: "Implementation[Node, Edge]",
        *,
        graph: Graph[str],
        data: dict[str, Attribute] | None = None,
    ) -> None: ...

    def __init__(
        self,
        *,
        graph: Graph[str],
        node_fn: Callable[[dict[str, Attribute]], Any] = Node,
        edge_fn: Callable[[dict[str, Attribute]], Any] = Edge,
        data: dict[str, Attribute] | None = None,
    ) -> None:
        """Create a new Implementation. Nodes and edges from `data` will not be automatically added to the graph.

        The constructor will combine `graph` and `data` so we can access nodes and edges of type `N` and `E` respectively.
        And use these to manipulate the underlying graph structure.
        It is the callers responsibility to keep `graph` and `data` in sync.
        This might seem like a limitation at first, but in fact it is a feature.
        It allows us to combine existing graph structures with new underlying data.
        A common use case for this is to handle subgraphs as in the following example:

        ```python
        subgraph = Graph()

        for node in original_impl.nodes.values():
            if fulfills_constraint(node):
                subgraph.add_node(node.name)

        for src, sink in original_impl.edges:
            if fulfills_constraint(original_impl.nodes[src]) and fulfills_constraint(original_impl.nodes[sink]):
                subgraph.add_edge(src, sink)

        new_impl = Implementation(graph=subgraph, data=original_impl.data)

        for node in new_impl.nodes.values():  # now we can iterate over the nodes of the subgraph
            do_something(node)
        ```
        """
        if data is None:
            data = {}
        if "nodes" not in data:
            data["nodes"] = {}
        if "edges" not in data:
            data["edges"] = {}
        super().__init__(data)

        self.graph = graph
        self._node_data: dict[str, Attribute] = cast(
            dict[str, Attribute], data["nodes"]
        )
        self._edge_data: dict[tuple[str, str], Attribute] = cast(
            dict[tuple[str, str], Attribute], data["edges"]
        )
        self._node_fn = node_fn
        self._edge_fn = edge_fn

    @overload
    def add_node(self, n: Node) -> Self: ...

    @overload
    def add_node(self, n: dict[str, Attribute]) -> Self: ...

    @overload
    def add_node(self, *, name: str, type: str, **attributes: Attribute) -> Self: ...

    def add_node(self, *args, **kwargs) -> Self:
        if len(args) == 1 and not kwargs:
            if isinstance(args[0], Node):
                return self.add_node(args[0].data)
            elif isinstance(args[0], dict):
                return self._add_node(args[0])
        elif len(args) == 0 and "name" in kwargs and "type" in kwargs:
            return self.add_node(kwargs)
        else:
            raise TypeError(
                "invalid arguments, expected either a Node or a dict or keyword arguments name, type, and more optional attributes"
            )
        return self

    def _add_node(self, n: dict[str, Attribute]) -> Self:
        self.graph.add_node(cast(str, n["name"]))
        self._node_data[cast(str, n["name"])] = n
        return self

    @overload
    def add_edge(self, e: Edge) -> Self: ...

    @overload
    def add_edge(self, e: dict) -> Self: ...

    @overload
    def add_edge(self, *, src: str, sink: str, **attributes: Attribute) -> Self: ...

    def add_edge(self, *args, **kwargs) -> Self:
        if len(args) == 1 and not kwargs:
            if isinstance(args[0], Edge):
                return self.add_edge(args[0].data)
            elif isinstance(args[0], dict):
                return self._add_edge(args[0])
        elif len(args) == 0 and "src" in kwargs and "sink" in kwargs:
            return self.add_edge(kwargs)
        raise TypeError(
            "invalid arguments, expected either an Edge or a dict or keyword arguments src, sink, and more optional attributes"
        )

    def _add_edge(self, e: dict) -> Self:
        self.graph.add_edge(e["src"], e["sink"])
        self._edge_data[(e["src"], e["sink"])] = e
        return self

    def add_nodes(self, ns: Iterable[Node | dict[str, Attribute]]) -> Self:
        for n in ns:
            self.add_node(n)
        return self

    def add_edges(self, es: Iterable[Edge | dict[str, Attribute]]) -> Self:
        for e in es:
            self.add_edge(e)
        return self

    def successors(self, node: str | Node) -> Mapping[str, N]:
        if not isinstance(node, str):
            node = node.name

        def _successors():
            return self.graph.successors[node]

        return self.get_node_mapping(_successors)

    def predecessors(self, node: str | Node) -> Mapping[str, N]:
        if not isinstance(node, str):
            node = node.name

        def _predecessors():
            return self.graph.predecessors[node]

        return self.get_node_mapping(_predecessors)

    def get_node_mapping(self, keys: Callable[[], Iterable[str]]) -> Mapping[str, N]:
        """Create a read-only mapping of names to Nodes in the order of `keys()`."""
        return _ReadOnlyMappingInOrderAsIterable(keys, self._node_data, self._node_fn)

    def get_edge_mapping(
        self, keys: Callable[[], Iterable[tuple[str, str]]]
    ) -> Mapping[tuple[str, str], E]:
        """Create a read-only mapping of (src, sink) pairs to Edges in the order of `keys()`."""
        return _ReadOnlyMappingInOrderAsIterable(keys, self._edge_data, self._edge_fn)

    @read_only_field
    def nodes(self, _: dict[str, Attribute]) -> Mapping[str, N]:
        return self.get_node_mapping(lambda: self.graph.nodes)

    @read_only_field
    def edges(self, _: dict[str, Attribute]) -> Mapping[tuple[str, str], E]:
        return self.get_edge_mapping(self.graph.iter_edges)

    def as_dict(self) -> dict[str, Attribute]:
        data = self.data.copy()
        data["nodes"] = cast(
            Attribute, list(n.data for n in self.nodes.values())
        )  # seems like mypy can't handle the nested recursive type hint list[dict[str, Attribute]]
        data["edges"] = cast(Attribute, list(e.data for e in self.edges.values()))
        return data

    def load_from_dict(
        self,
        d: dict[str, Any],
    ) -> "Implementation[N, E]":
        """Load attributes, nodes and edges from a dictionary, that was created by `as_dict`.

        Opposed to the constructor, this will add all nodes and edges to the underlying graph.
        :::{important}
        This will override all state of the current object.
        :::
        """
        data = d.copy()
        nodes = {node["name"]: node for node in data["nodes"]}
        edges = {(edge["src"], edge["sink"]): edge for edge in data["edges"]}
        data["nodes"] = nodes
        data["edges"] = edges
        for node in nodes.values():
            self.add_node(node)
        for edge in edges.values():
            self.add_edge(edge)
        self.data = data
        return self


_K = TypeVar("_K")
_V = TypeVar("_V")


class _ReadOnlyMappingInOrderAsIterable(Mapping[_K, _V]):
    def __init__(
        self,
        iterable: Callable[[], Iterable[_K]],
        d: dict[_K, Any],
        value_constructor: Callable[[Any], _V],
    ):
        self._iterable = iterable
        self._d = d
        self._value_constructor = value_constructor

    def __iter__(self) -> Iterator[_K]:
        yield from self._iterable()

    def __len__(self) -> int:
        return len(self._d)

    def __contains__(self, k: object) -> bool:
        return k in self._d

    def __getitem__(self, k: _K) -> _V:
        return self._value_constructor(self._d[k])

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Mapping):
            return NotImplemented
        return dict(self) == dict(other)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({dict(self)})"



from elasticai.creator.nn.quantized_grads.base_modules import Conv2d
from elasticai.creator.nn.quantized_grads.fixed_point import (
    FixedPointConfigV2,
    QuantizeForwHTE,
    QuantizeParamToFixedPointHTE,
    QuantizeParamToFixedPointStochastic,
)


def test_conv1d_fxp_init():
    conf = FixedPointConfigV2(8, 3)

    Conv2d(
        QuantizeForwHTE(conf),
        QuantizeParamToFixedPointHTE(conf),
        3,
        2,
        2,
        bias_quantization=QuantizeParamToFixedPointStochastic(conf),
    )



from collections.abc import Callable, Iterable
from typing import Any

import torch.nn as nn
from torch import fx

from elasticai.creator.ir import LoweringPass
from elasticai.creator.torch2ir import Implementation

from .default_handlers import handlers


class Ir2Torch(LoweringPass[Implementation, nn.Module]):
    def convert(
        self, ir: Iterable[Implementation], state_dict: dict[str, Any] | None = None
    ) -> nn.Module:
        """Rebuild the original pytorch model from a given IR.

        :param: `ir`: You need to make sure that `Ir2Torch` has a type handler for each implementation in `ir`
        :param: `state_dict`: You can optionally pass a state dict. This should be a state dict created
            from the original model via `nn.Module.state_dict`. As the `Torch2Ir` stage got rid of all
            duplicate submodules, we will strip all unknown keys from the `state_dict` and then load it.
        """
        root_module = nn.Module()
        root = ""
        _ir = {impl.name: impl for impl in ir}
        for impl in _ir.values():
            if impl.type != "module":
                modules = list(self([impl]))
                assert len(modules) == 1
                root_module.add_module(impl.name, modules[0])
            else:
                root = impl.name

        graph = fx.Graph()
        ir_root = _ir[root]
        for ir_node in ir_root.nodes.values():
            if ir_node.type not in ("input", "output"):
                graph.call_module(
                    ir_node.implementation, tuple(ir_root.predecessors(ir_node))
                )

        module = fx.GraphModule(root_module, graph)

        if state_dict is not None:
            filtered_state_dict = {}
            for key, value in state_dict.items():
                submodule = ".".join(key.split(".")[:-1])
                if submodule in _ir:
                    filtered_state_dict[key] = value
            module.load_state_dict(filtered_state_dict)
        return module

    def register_type_handlers(
        self, handlers: Iterable[Callable[[Implementation], nn.Module]]
    ) -> None:
        for handler in handlers:
            self.register(handler.__name__)(handler)


def get_default_converter() -> Ir2Torch:
    converter = Ir2Torch()
    converter.register_type_handlers(handlers)
    return converter



from .utils import build_graph_from_dict, find_matches

"""
Test List:

 * if adding a function to find all subgraphs, we need to check that
   matches already found are not discarded again. Because the implementation
   will try to visit matches that are already completed, which might cause
   them to be removed from the next list of potential matches.

"""


def test_find_identical_graph_with_single_node():
    g = build_graph_from_dict(
        {
            ("a", "a_t"): [],
        }
    )
    p = build_graph_from_dict(
        {
            ("a", "a_t"): [],
        }
    )
    assert find_matches(g, p) == [{"a": "a"}]


def test_find_identical_graph_with_multiple_nodes():
    g = build_graph_from_dict(
        {
            ("a", "a_t"): ["b"],
            ("b", "b_t"): [],
        }
    )
    p = build_graph_from_dict(
        {
            ("a", "a_t"): ["b"],
            ("b", "b_t"): [],
        }
    )
    assert find_matches(g, p) == [{"a": "a", "b": "b"}]


def test_find_graph_by_type():
    g = build_graph_from_dict(
        {
            ("a", "a_t"): ["b"],
            ("b", "b_t"): [],
        }
    )
    p = build_graph_from_dict(
        {
            ("0", "a_t"): ["1"],
            ("1", "b_t"): [],
        }
    )
    assert find_matches(g, p) == [{"0": "a", "1": "b"}]


def test_find_aba_patter_two_times():
    g = build_graph_from_dict(
        {
            ("a", "a_t"): ["b", "c"],
            ("b", "b_t"): ["d"],
            ("c", "c_t"): ["d"],
            ("d", "a_t"): ["e"],
            ("e", "b_t"): [],
        }
    )

    pattern = build_graph_from_dict(
        {
            ("0", "a_t"): ["1"],
            ("1", "b_t"): [],
        }
    )

    matches = find_matches(g, pattern)

    assert matches == [
        {"0": "d", "1": "e"},
        {"0": "a", "1": "b"},
    ]


def test_find_pattern_if_we_have_to_search_for_predecessors():
    g = build_graph_from_dict(
        {("a", "a_t"): ["c"], ("b", "b_t"): ["c"], ("c", "c_t"): []}
    )
    p = build_graph_from_dict(
        {("0", "a_t"): ["2"], ("1", "b_t"): ["2"], ("2", "c_t"): []}
    )
    assert find_matches(g, p) == [{"0": "a", "1": "b", "2": "c"}]



from collections.abc import Callable, Iterator
from functools import singledispatchmethod
from typing import Generic, Protocol, TypeVar, overload

from elasticai.creator.graph.graph import Graph

T = TypeVar("T", int, str)
TPattern = TypeVar("TPattern", int, str)

_Tcon = TypeVar("_Tcon", contravariant=True)
_TPatternCon = TypeVar("_TPatternCon", contravariant=True)


class NodeConstraint(Protocol[_TPatternCon, _Tcon]):
    def __call__(self, *, pattern_node: _TPatternCon, graph_node: _Tcon) -> bool: ...


def find_subgraphs(
    *,
    pattern: Graph[TPattern],
    graph: Graph[T],
    node_constraint: NodeConstraint[TPattern, T],
) -> list[dict[TPattern, T]]:
    """Find occurences of `pattern` in `graph` where nodes match according to `node_constraint`.

    Use the `node_constraint` function to define how nodes in `graph` and `pattern` should match.
    Even though the function only takes the node identifiers as arguments, you can still compare more complex objects, e.g.,
    as follows:

    ```python
    graph_nodes_data = {
        "a": {"type": "A", "channels": 1},
        "b": {"type": "B", "channels": 2},
    }
    pattern_nodes_data = {
        "0": {"type": "A", "channels": 1},
        "1": {"type": "B", "channels": 2},
    }

    def node_constraint(pattern_node: str, graph_node: str) -> bool:
        return graph_nodes_data[graph_node]["type"] == pattern_nodes_data[pattern_node]["type"]

    subgraphs = find_subgraphs(graph, pattern, node_constraint)
    ```

    :::{warning}
    This function is not optimized nor do we have benchmarks for performance. You might run into situations where runtime is too long for your use case.
    :::

    :::{important}
    The function will stop as soon as a match was found. That does not mean, that it will return at most one match.
    Instead matches that were found in the same iteration are returned together.
    :::

    :param: `graph`: The graph to search in.
    :param: `pattern`: Defines the pattern to search for. A node in the pattern will match if
       1. it has fewer or equal in- and out-edges as the corresponding `graph` node
       2. `node_constraint` holds for the corresponding node in `graph`
        pattern (Graph): _description_
       3. 1. and 2. hold for all predecessors and successors of the node
    :param: `node_constraint`: An additional constraint that must hold for a node in `graph` to match a node in `pattern`. If you just want to match the graph structure, independent of any node attributes, you can pass `lambda a, b: True` as the constraint.

    :return: A list of dictionaries, where each dictionary maps the names of the nodes in `pattern` to the names of the corresponding nodes in `graph`. The list contains one dictionary for each match found.
    """

    def stop_condition(matches: list[_Match[TPattern, T]]) -> bool:
        return any(m.is_complete() for m in matches) or len(matches) == 0

    return _SubgraphMatcher(
        pattern, graph, node_constraint, stop_condition=stop_condition
    ).find_matches()


class _MatcherNode(Generic[T]):
    def __init__(self, name: T, owner: Graph[T]):
        self.name: T = name
        self.owner: Graph[T] = owner

    def __hash__(self) -> int:
        return hash(self.name)

    @property
    def in_degree(self) -> int:
        return len(self.owner.predecessors[self.name])

    @property
    def out_degree(self) -> int:
        return len(self.owner.successors[self.name])

    @property
    def successors(self) -> Iterator["_MatcherNode[T]"]:
        for node in self.owner.successors[self.name]:
            yield _MatcherNode(node, self.owner)

    @property
    def predecessors(self) -> Iterator["_MatcherNode[T]"]:
        for node in self.owner.predecessors[self.name]:
            yield _MatcherNode(node, self.owner)


class _Match(Generic[TPattern, T]):
    def __init__(
        self,
        last_pattern_node: TPattern | None,
        pattern: Graph[TPattern],
        graph: Graph[T],
        map: dict[TPattern, T] | None = None,
    ):
        if map is None:
            map = {}
        self.map: dict[TPattern, T] = map
        self.pattern: Graph[TPattern] = pattern
        self.graph: Graph[T] = graph
        self._last_pattern_node: TPattern | None = last_pattern_node

    def _get_last_pattern_node(self) -> _MatcherNode[TPattern]:
        assert self._last_pattern_node is not None
        return _MatcherNode(self._last_pattern_node, self.pattern)

    @property
    def pattern_successors(self) -> Iterator[_MatcherNode[TPattern]]:
        return self._get_last_pattern_node().successors

    @property
    def graph_successors(self) -> Iterator[_MatcherNode[T]]:
        return self._last_graph_node().successors

    @property
    def pattern_predecessors(self) -> Iterator[_MatcherNode[TPattern]]:
        return self._get_last_pattern_node().predecessors

    @property
    def graph_predecessors(self) -> Iterator[_MatcherNode[T]]:
        return self._last_graph_node().predecessors

    def _last_graph_node(self) -> _MatcherNode[T]:
        assert self._last_pattern_node is not None
        return _MatcherNode(self.map[self._last_pattern_node], self.graph)

    @singledispatchmethod
    def visited_pattern(self, pattern_node) -> bool:
        return self.visited_pattern(_MatcherNode(pattern_node, self.pattern))

    @visited_pattern.register
    def _(self, pattern_node: _MatcherNode) -> bool:
        return pattern_node.name in self.map

    @singledispatchmethod
    def visited_graph(self, graph_node) -> bool:
        return self.visited_graph(_MatcherNode(graph_node, self.graph))

    @visited_graph.register
    def _(self, graph_node: _MatcherNode) -> bool:
        return graph_node.name in self.map.values()

    @overload
    def visited_neither(self, pattern_node: T, graph_node: T) -> bool: ...

    @overload
    def visited_neither(
        self,
        pattern_node: _MatcherNode[TPattern],
        graph_node: _MatcherNode[T],
    ) -> bool: ...

    @singledispatchmethod  # type: ignore[operator, misc]
    def visited_neither(self, pattern_node, graph_node) -> bool:
        return self.visited_neither(
            _MatcherNode(pattern_node, self.pattern),
            _MatcherNode(graph_node, self.graph),
        )

    @visited_neither.register  # type: ignore[attr-defined]
    def _(
        self,
        pattern_node: _MatcherNode,
        graph_node: _MatcherNode,
    ) -> bool:
        return not (
            self.visited_pattern(pattern_node) or self.visited_graph(graph_node)
        )

    @overload
    def visit(self, pattern_node: TPattern, graph_node: T) -> "_Match[T, TPattern]": ...

    @overload
    def visit(
        self,
        pattern_node: _MatcherNode[TPattern],
        graph_node: _MatcherNode[T],
    ) -> "_Match[TPattern, T]": ...

    @singledispatchmethod  # type: ignore[operator, misc]
    def visit(self, pattern_node, graph_node):
        return self.visit(
            _MatcherNode(pattern_node, self.pattern),
            _MatcherNode(graph_node, self.graph),
        )

    @visit.register  # type: ignore[attr-defined]
    def _(
        self,
        pattern_node: _MatcherNode,
        graph_node: _MatcherNode,
    ):
        return _Match(
            pattern_node.name,
            self.pattern,
            self.graph,
            self.map | dict(((pattern_node.name, graph_node.name),)),
        )

    def is_complete(self) -> bool:
        return len(self.map) == len(self.pattern.nodes)


class _SubgraphMatcher(Generic[TPattern, T]):
    def __init__(
        self,
        pattern: Graph[TPattern],
        graph: Graph[T],
        node_constraint: NodeConstraint[TPattern, T],
        stop_condition: Callable[[list[_Match[TPattern, T]]], bool],
    ):
        self._graph: Graph[T] = graph
        self._pattern: Graph[TPattern] = pattern
        self._node_constraint: NodeConstraint[TPattern, T] = node_constraint
        self._stop_condition: Callable[[list[_Match[TPattern, T]]], bool] = (
            stop_condition
        )

    def find_matches(self) -> list[dict[TPattern, T]]:
        potential_matches = self._build_initial_potential_matches()
        while not self._stop_condition(potential_matches):
            potential_matches = self._build_next_potential_matches(potential_matches)
        matches = [pm.map for pm in potential_matches if pm.is_complete()]
        return matches

    def _build_initial_potential_matches(
        self,
    ) -> list[_Match[TPattern, T]]:
        pn = self._pick_a_start_node_from_pattern()
        gns = []
        for name in self._find_all_start_nodes_from_graph(pn.name):
            gns.append(_MatcherNode(name, self._graph))

        potential_matches: list[_Match[TPattern, T]] = []
        base_element = _Match(None, self._pattern, self._graph)
        for gn in gns:
            potential_matches.append(base_element.visit(pn, gn))

        return potential_matches

    def _build_next_potential_matches(
        self, potential_matches: list[_Match[TPattern, T]]
    ) -> list[_Match[TPattern, T]]:
        next_potential_matches: list[_Match[TPattern, T]] = []
        for potential_match in potential_matches:
            next_potential_matches.extend(
                self._find_follow_ups_for_match(potential_match)
            )
        return next_potential_matches

    def _find_follow_ups_for_match(
        self, potential_match: _Match[TPattern, T]
    ) -> list[_Match[TPattern, T]]:
        next_potential_matches: list[_Match[TPattern, T]] = []

        def should_visit(pn, gn):
            return potential_match.visited_neither(pn, gn) and self._nodes_match(pn, gn)

        def visit(pn, gn):
            next_potential_matches.append(potential_match.visit(pn, gn))

        for ps in potential_match.pattern_successors:
            for gs in potential_match.graph_successors:
                if should_visit(ps, gs):
                    visit(ps, gs)

        for pp in potential_match.pattern_predecessors:
            for gp in potential_match.graph_predecessors:
                if should_visit(pp, gp):
                    visit(pp, gp)

        return next_potential_matches

    @overload
    def _nodes_match(self, pattern_node: TPattern, graph_node: T) -> bool: ...

    @overload
    def _nodes_match(
        self,
        pattern_node: _MatcherNode[TPattern],
        graph_node: _MatcherNode[T],
    ) -> bool: ...

    @singledispatchmethod  # type: ignore[operator, misc]
    def _nodes_match(
        self, pattern_node: _MatcherNode[TPattern], graph_node: _MatcherNode[T]
    ) -> bool:
        return self._node_constraint(graph_node.name, pattern_node.name)

    # ignoring some errors as mypy can't handle overload+singledispatch
    @singledispatchmethod  # type: ignore[operator, misc]
    def _nodes_match(self, pattern_node, graph_node) -> bool:
        return self._nodes_match(
            _MatcherNode(pattern_node, self._pattern),
            _MatcherNode(graph_node, self._graph),
        )

    @_nodes_match.register  # type: ignore[attr-defined, misc]
    def _nodes_match(
        self, pattern_node: _MatcherNode, graph_node: _MatcherNode
    ) -> bool:
        return self._node_constraint(
            graph_node=graph_node.name, pattern_node=pattern_node.name
        )

    def _pick_a_start_node_from_pattern(self) -> _MatcherNode[TPattern]:
        return _MatcherNode(next(iter(self._pattern.nodes)), self._pattern)

    def _find_all_start_nodes_from_graph(self, start_node: TPattern) -> set[T]:
        matches = set()
        for node in self._graph.nodes:
            if self._node_constraint(pattern_node=start_node, graph_node=node):
                matches.add(node)
        return matches



from elasticai.creator.graph import BaseGraph as GraphDelegate
from elasticai.creator.graph import bfs_iter_up, dfs_pre_order


def test_yield_node_in_case_it_has_no_successors():
    g = GraphDelegate.from_dict(
        {
            "0": [],
        }
    )
    actual = tuple(g.iter_nodes())
    assert actual == ("0",)


def test_iterating_breadth_first_upwards():
    g = GraphDelegate()
    """
             0
             |
          /-----\
          |     |
          1     2
          | /---+
          |/    |
          3     4
          |     |
          |     6
          |/----+
          5
    """
    g = GraphDelegate.from_dict(
        {
            "0": ["1", "2"],
            "1": ["3"],
            "2": ["3", "4"],
            "3": ["5"],
            "4": ["6"],
            "6": ["5"],
        }
    )

    actual = tuple(bfs_iter_up(g.get_predecessors, g.get_successors, "5"))
    assert actual == ("3", "6", "1", "4", "2", "0")


def test_iterating_depth_first_preorder():
    g = GraphDelegate()
    """
             0
             |
          /-----\
          |     |
          1     2
          | /---+
          |/    |
          3     4
          |     |
          |     6
          |/----+
          5
    """
    g = GraphDelegate.from_dict(
        {
            "0": ["1", "2"],
            "1": ["3"],
            "2": ["3", "4"],
            "3": ["5"],
            "4": ["6"],
            "6": ["5"],
        }
    )

    actual = tuple(dfs_pre_order(g.get_successors, "0"))
    expected = ("0", "1", "3", "5", "2", "4", "6")
    assert actual == expected



from pathlib import Path

import pytest
import torch

from elasticai.creator.file_generation.on_disk_path import OnDiskPath
from elasticai.creator.nn.fixed_point.linear.layer.linear import Linear
from elasticai.creator.vhdl.ghdl_simulation import GHDLSimulator
from elasticai.creator.vhdl.simulated_layer import SimulatedLayer


def create_ones_input_list(batch_size: int, in_feature_num: int):
    return [[[1.0] * in_feature_num]] * batch_size


@pytest.mark.simulation
@pytest.mark.parametrize(
    "x", ([[[0.0, 1.0, 1.0]]], [[[1.0, 1.0, 1.0]]], [[[2.0, 1.0, 0.0]]])
)
def test_verify_hw_sw_equivalence_3_inputs(
    x: list[list[list[float]]], tmp_path: Path
) -> None:
    input_data = torch.Tensor(x)
    sw_conv = Linear(
        in_features=3,
        out_features=2,
        total_bits=4,
        frac_bits=1,
        bias=True,
    )
    sw_conv.weight.data = torch.ones_like(sw_conv.weight)
    sw_conv.bias.data = torch.ones_like(sw_conv.bias)

    sw_output = sw_conv(input_data)

    design = sw_conv.create_design("linear")
    testbench = sw_conv.create_testbench("linear_testbench", design)

    build_dir = OnDiskPath(name=tmp_path.name, parent=str(tmp_path.parent))
    design.save_to(build_dir.create_subpath("srcs"))
    testbench.save_to(build_dir.create_subpath("testbenches"))

    sim_layer = SimulatedLayer(testbench, GHDLSimulator, working_dir=tmp_path)
    sim_output = sim_layer(input_data)

    assert sw_output.tolist() == sim_output


@pytest.mark.simulation
@pytest.mark.parametrize(
    "x",
    (
        create_ones_input_list(2, 4),
        [[[0.5, 0.25, -1.0, 1.0]], [[-1.0, 1.0, -1.0, 1.0]]],
        [[[0.0, 1.0, 1.0, 0.0]], [[-1.0, 1.0, -1.0, 1.0]]],
    ),
)
def test_verify_hw_sw_equivalence_4_inputs(
    x: list[list[list[float]]], tmp_path: Path
) -> None:
    input_data = torch.Tensor(x)
    sw_conv = Linear(
        in_features=4,
        out_features=10,
        total_bits=4,
        frac_bits=1,
        bias=True,
    )
    sw_conv.weight.data = torch.ones_like(sw_conv.weight)
    sw_conv.bias.data = torch.ones_like(sw_conv.bias)

    sw_output = sw_conv(input_data)

    design = sw_conv.create_design("linear")
    testbench = sw_conv.create_testbench("linear_testbench", design)

    build_dir = OnDiskPath(name=tmp_path.name, parent=str(tmp_path.parent))
    design.save_to(build_dir.create_subpath("srcs"))
    testbench.save_to(build_dir.create_subpath("testbenches"))

    sim_layer = SimulatedLayer(testbench, GHDLSimulator, working_dir=tmp_path)
    sim_output = sim_layer(input_data)

    assert sw_output.tolist() == sim_output


@pytest.mark.simulation
@pytest.mark.parametrize(
    "x",
    (
        [
            [[0.0, 0.0, 0.0, 0.0]],
            [[1.0, 0.0, 0.0, 0.0]],
            [[0.0, 1.0, 0.0, 0.0]],
            [[0.0, 0.0, 1.0, 0.0]],
            [[0.0, 0.0, 0.0, 1.0]],
            [[2.0, 0.0, 0.0, 0.0]],
            [[1.0, 1.0, 0.0, 0.0]],
            [[1.0, 0.0, 1.0, 0.0]],
            [[1.0, 0.0, 0.0, 1.0]],
            [[3.0, 0.0, 0.0, 0.0]],
            [[1.0, 2.0, 0.0, 0.0]],
            [[1.0, 0.0, 2.0, 0.0]],
            [[1.0, 0.0, 0.0, 2.0]],
            [[1.0, 1.0, 0.0, 0.0]],
            [[0.0, 1.0, 1.0, 0.0]],
            [[0.0, 0.0, 1.0, 1.0]],
            [[1.0, 0.0, 0.0, 1.0]],
            [[-1.0, 0.0, 0.0, 0.0]],
            [[0.0, -1.0, 0.0, 0.0]],
            [[0.0, 0.0, -1.0, 0.0]],
            [[0.0, 0.0, 0.0, -1.0]],
            [[2.0, 0.0, 0.0, 0.0]],
            [[0.0, 2.0, 0.0, 0.0]],
            [[0.0, 0.0, 2.0, 0.0]],
            [[0.0, 0.0, 0.0, 2.0]],
        ],
    ),
)
def test_verify_hw_sw_equivalence_4_inputs_3_outputs(
    x: list[list[list[float]]], tmp_path: Path
) -> None:
    input_data = torch.Tensor(x)
    sw_conv = Linear(
        in_features=4,
        out_features=3,
        total_bits=8,
        frac_bits=2,
        bias=True,
    )
    sw_conv.weight.data = torch.ones_like(sw_conv.weight) * 2
    sw_conv.bias.data = torch.Tensor([1.0, 2.0, -1.0])

    sw_output = sw_conv(input_data)

    design = sw_conv.create_design("linear")
    testbench = sw_conv.create_testbench("linear_testbench", design)

    build_dir = OnDiskPath(name=tmp_path.name, parent=str(tmp_path.parent))
    design.save_to(build_dir.create_subpath("srcs"))
    testbench.save_to(build_dir.create_subpath("testbenches"))

    sim_layer = SimulatedLayer(testbench, GHDLSimulator, working_dir=tmp_path)
    sim_output = sim_layer(input_data)

    assert sw_output.tolist() == sim_output



from elasticai.creator.ir2vhdl import Instance, VhdlNode

from .clocked_combinatorial import ClockedInstance
from .node_factory import InstanceFactoryForCombinatorial


@InstanceFactoryForCombinatorial.register
def shift_register(node: VhdlNode) -> Instance:
    return ShiftRegister(node)


class ShiftRegister(ClockedInstance):
    _logic_signals_with_default_suffix = ("valid_in", "valid_out")

    def __init__(self, node: VhdlNode):
        data_width = node.input_shape.size()
        num_points = node.output_shape.width
        output_width = node.output_shape.size()
        super().__init__(
            node,
            input_width=data_width,
            output_width=output_width,
            generic_map={"data_width": data_width, "num_points": num_points},
        )



from elasticai.creator.graph import BaseGraph as _Graph
from elasticai.creator.ir import Edge, Implementation, Node, edge, node
from elasticai.creator.ir.base.attribute import Attribute


class Graph(Implementation[Node, Edge]):
    name: str
    type: str

    def __init__(self, data: dict[str, Attribute] | None = None):
        g: _Graph[str] = _Graph()
        if data is None:
            data = {}
        super().__init__(
            graph=g,
            node_fn=Node,
            edge_fn=Edge,
            data=data,
        )


def test_graph_is_serialized():
    g = Graph(data=dict(name="network", type="network"))
    g.add_node(node(name="node1", type="type1"))
    g.add_node(node(name="node2", type="type2"))
    g.add_edge(edge(src="node1", sink="node2"))
    assert g.data == {
        "name": "network",
        "type": "network",
        "nodes": {
            "node1": {"name": "node1", "type": "type1"},
            "node2": {"name": "node2", "type": "type2"},
        },
        "edges": {
            ("node1", "node2"): {"src": "node1", "sink": "node2"},
        },
    }


def test_graph_has_required_fields():
    g = Graph(data=dict(name="network", type="network"))
    assert g.name == "network"



import pytest
import torch
from torch.nn.functional import silu as torch_silu

from elasticai.creator.nn.fixed_point import quantize as fxp_quantize
from elasticai.creator.nn.fixed_point.precomputed.adaptable_silu import AdaptableSiLU
from tests.tensor_test_case import assertTensorEqual


def fxp_args() -> dict[str, int]:
    return dict(total_bits=16, frac_bits=8)


def quantize(x: torch.Tensor) -> torch.Tensor:
    return fxp_quantize(x=x, **fxp_args())


def forge_silu(scale: float, beta: float) -> AdaptableSiLU:
    silu = AdaptableSiLU(num_steps=5, sampling_intervall=(-10, 10), **fxp_args())
    silu.load_state_dict(
        {
            "_base_module.scale": torch.tensor([scale]),
            "_base_module.beta": torch.tensor([beta]),
            "_step_lut": torch.linspace(-10, 10, 5),
        }
    )
    return silu


@pytest.mark.parametrize("scale,beta", [(1, 0), (-1, 0), (1, 1), (1, -1), (2, 3)])
def test_silu_with_several_scale_and_beta(scale: float, beta):
    silu = forge_silu(scale, beta)

    inputs = torch.tensor([-15.0, -10.0, -5.0, 0.0, 5.0, 10.0, 15.0])
    actual_outputs = silu(inputs)

    target_inputs = torch.tensor([-10.0, -10.0, -5.0, 0.0, 5.0, 10.0, 10.0])
    target_outputs = quantize(scale * torch_silu(target_inputs) + beta)

    assertTensorEqual(target_outputs, actual_outputs)



from typing import cast

from elasticai.creator.file_generation.in_memory_path import InMemoryFile, InMemoryPath
from elasticai.creator.nn.fixed_point.relu.layer import ReLU


def test_vhdl_code_matches_expected() -> None:
    expected = """-- This is the ReLU implementation for fixed-point data
-- it only checks the highest bit of the input data
-- when the CLOCK_OPTION is enabled, please notice the data only updates until the clock arises.
-- Version: 1.0
-- Created by: Chao
-- Last modified date: 2022.11.06

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity relu is
    generic (
        DATA_WIDTH   : integer := 16;
        CLOCK_OPTION : boolean := true
    );
    port (
        enable : in std_logic;
    	clock  : in std_logic;
    	x      : in std_logic_vector(DATA_WIDTH-1 downto 0);
    	y      : out std_logic_vector(DATA_WIDTH-1 downto 0)
    );
end entity relu;

architecture rtl of relu is
    signal fxp_input : signed(DATA_WIDTH-1 downto 0) := (others=>'0');
    signal fxp_output : signed(DATA_WIDTH-1 downto 0) := (others=>'0');
begin
    fxp_input <= signed(x);
    y <= std_logic_vector(fxp_output);

    clocked: if CLOCK_OPTION generate
        main_process : process (enable, clock)
        begin
            if (enable = '0') then
                fxp_output <= to_signed(0, DATA_WIDTH);
            elsif (rising_edge(clock)) then

                if fxp_input < 0 then
                    fxp_output <= to_signed(0, DATA_WIDTH);
                else
                    fxp_output <= fxp_input;
                end if;
            end if;
        end process;
    end generate;

    async: if (not CLOCK_OPTION) generate
        process (enable, fxp_input)
        begin
            if enable = '0' then
                fxp_output <= to_signed(0, DATA_WIDTH);
            else
                if fxp_input < 0 then
                    fxp_output <= to_signed(0, DATA_WIDTH);
                else
                    fxp_output <= fxp_input;
                end if;
            end if;
        end process;
    end generate;
end architecture rtl;
""".splitlines()
    relu = ReLU(total_bits=16, use_clock=True)
    build_path = InMemoryPath("build", parent=None)
    design = relu.create_design("relu")
    design.save_to(build_path)
    actual = cast(InMemoryFile, build_path["relu"]).text
    assert actual == expected



import pytest

from elasticai.creator.vhdl.code_generation.code_abstractions import (
    to_vhdl_binary_string,
)


def test_to_vhdl_binary_string_raises_error_if_value_not_representable() -> None:
    with pytest.raises(ValueError):
        _ = to_vhdl_binary_string(-8, 3)


@pytest.mark.parametrize(
    ("number", "number_of_bits", "expected"),
    [
        (0, 1, '"0"'),
        (1, 1, '"1"'),
        (6, 4, '"0110"'),
        (-6, 4, '"1010"'),
        (7, 5, '"00111"'),
        (-7, 5, '"11001"'),
    ],
)
def test_to_vhdl_binary_string(number: int, number_of_bits: int, expected: str) -> None:
    assert to_vhdl_binary_string(number, number_of_bits) == expected



from typing import Protocol, TypeVar

T = TypeVar("T")


class IrData(Protocol[T]):
    data: dict[str, dict[str, T]]


EdgeT = TypeVar("EdgeT", bound=IrData)

PNodeT = TypeVar("PNodeT", bound=IrData)
INodeT = TypeVar("INodeT", bound=IrData)
RNodeT = TypeVar("RNodeT", bound=IrData)
GNodeT = TypeVar("GNodeT", bound=IrData)



from dataclasses import dataclass
from typing import cast
from unittest import TestCase, main

import torch
from torch import Tensor, cat, tensor, testing, zeros
from torch.nn import BatchNorm1d, Linear, Sequential

from elasticai.creator.file_generation.in_memory_path import InMemoryFile, InMemoryPath
from elasticai.creator.nn import Sequential as SequentialCreator
from elasticai.creator.nn.fixed_point import (
    BatchNormedLinear as BatchNormedLinearCreator,
)
from elasticai.creator.nn.fixed_point import (
    Linear as LinearCreator,
)


@dataclass
class SettingsTest:
    data_repeat_num: int = 3
    size_input_layer: int = 3
    size_output_layer: int = 1
    quant_width_total: int = 12
    quant_width_frac: int = 8


RecommendedSettings = SettingsTest()


def generate_test_data(settings: SettingsTest = RecommendedSettings) -> Tensor:
    outputs = zeros((1, settings.size_input_layer)).repeat(settings.data_repeat_num, 1)
    for idx in range(settings.data_repeat_num):
        outputs = cat((outputs, tensor([[-1.5, 0.5, 1.5]])), dim=0)
        outputs = cat((outputs, tensor([[1.5, 0.5, -1.5]])), dim=0)
    return outputs


def generate_torch_linear(settings: SettingsTest = RecommendedSettings) -> list:
    linear_layer = Linear(
        settings.size_input_layer, settings.size_output_layer, bias=True
    )
    linear_layer.bias.data = tensor(0.5)
    linear_layer.weight.data = tensor([[0.5, 1.0, -0.5]])
    return [linear_layer]


def generate_creator_linear(settings: SettingsTest = RecommendedSettings) -> list:
    linear_layer = LinearCreator(
        in_features=settings.size_input_layer,
        out_features=settings.size_output_layer,
        total_bits=settings.quant_width_total,
        frac_bits=settings.quant_width_frac,
        bias=True,
    )
    linear_layer.bias.data = tensor(0.5)
    linear_layer.weight.data = tensor([[0.5, 1.0, -0.5]])
    return [linear_layer]


def generate_torch_batchlinear(settings: SettingsTest = RecommendedSettings) -> list:
    linear_layer = generate_torch_linear(settings)

    batch1d_layer = BatchNorm1d(settings.size_output_layer)
    batch1d_layer.bias.data = tensor([0.25])
    batch1d_layer.weight.data = tensor([0.75])
    linear_layer.append(batch1d_layer)
    return linear_layer


def generate_creator_batchlinear(settings: SettingsTest = RecommendedSettings) -> list:
    batch1d_layer = BatchNormedLinearCreator(
        in_features=settings.size_input_layer,
        out_features=settings.size_output_layer,
        total_bits=settings.quant_width_total,
        frac_bits=settings.quant_width_frac,
        bias=True,
        bn_affine=True,
    )
    batch1d_layer.lin_bias.data = tensor([[0.5]])
    batch1d_layer.lin_weight.data = tensor([[0.5, 1.0, -0.5]])

    batch1d_layer.bn_bias.data = tensor([0.25])
    batch1d_layer.bn_weight.data = tensor([0.75])
    return [batch1d_layer]


def build_torch_sequential(nn_layer: list, do_training: bool = False) -> Sequential:
    model_build = Sequential()
    for layer in nn_layer:
        model_build.append(layer)

    if do_training:
        model_build.train()
    else:
        model_build.eval()
    return model_build


def build_creator_sequential(
    nn_layer: list, do_training: bool = False
) -> SequentialCreator:
    model_build = SequentialCreator()
    for layer in nn_layer:
        model_build.append(layer)

    if do_training:
        model_build.train()
    else:
        model_build.eval()
    return model_build


class TestCreatorLinear(TestCase):
    model_torch = build_torch_sequential(generate_torch_linear())
    model_creator = build_creator_sequential(generate_creator_linear())
    data_in = generate_test_data()

    def test_data_input_size(self):
        testing.assert_close(self.data_in.size(1), RecommendedSettings.size_input_layer)

    def test_data_output_size_torch(self):
        testing.assert_close(
            self.model_torch(self.data_in).size(1),
            RecommendedSettings.size_output_layer,
        )

    def test_data_output_size_creator(self):
        testing.assert_close(
            self.model_creator(self.data_in).size(1),
            RecommendedSettings.size_output_layer,
        )

    def test_creator_linear(self):
        data_out_torch = self.model_torch(self.data_in)
        data_out_creator = self.model_creator(self.data_in)
        testing.assert_close(
            data_out_torch, data_out_creator, atol=1 / (2**8), rtol=1 / (2**8)
        )


def test_inference_of_multidimensional_data() -> None:
    linear = LinearCreator(
        total_bits=16, frac_bits=8, in_features=3, out_features=2, bias=False
    )
    linear.weight.data = torch.ones_like(linear.weight.data)

    inputs = torch.tensor([1.0, 2.0, 3.0])
    expected = [6.0, 6.0]
    actual = linear(inputs).tolist()

    assert expected == actual


def test_overflow_behaviour() -> None:
    linear = LinearCreator(
        total_bits=4, frac_bits=1, in_features=2, out_features=1, bias=False
    )
    linear.weight.data = torch.ones_like(linear.weight.data) * 2

    inputs = torch.tensor([2.5, -1.0])
    expected = [3.0]  # quantize(2.5 * 2 - 1.0 * 2)
    actual = linear(inputs).tolist()

    assert expected == actual


def test_underflow_behaviour() -> None:
    linear = LinearCreator(
        total_bits=4, frac_bits=1, in_features=1, out_features=1, bias=False
    )
    linear.weight.data = torch.ones_like(linear.weight.data) * 0.5

    inputs = torch.tensor([0.5])
    expected = [0.0]
    actual = linear(inputs).tolist()

    assert expected == actual


def test_bias_addition() -> None:
    linear = LinearCreator(
        total_bits=16, frac_bits=8, in_features=1, out_features=1, bias=True
    )
    linear.weight.data = torch.ones_like(linear.weight.data)
    linear.bias.data = torch.ones_like(linear.bias.data) * 2

    inputs = torch.tensor([3.0])
    expected = [5.0]
    actual = linear(inputs).tolist()

    assert expected == actual


def test_linear_layer_creates_correct_design() -> None:
    expected_linear_code = """library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;               -- for type conversions

library work;
use work.all;

entity linear is -- layer_name is for distinguish same type of layers (with various weights) in one module
    generic (
        DATA_WIDTH   : integer := 16;
        FRAC_WIDTH   : integer := 8;
        X_ADDR_WIDTH : integer := 2;
        Y_ADDR_WIDTH : integer := 1;
        IN_FEATURE_NUM : integer := 3;
        OUT_FEATURE_NUM : integer := 2;
        RESOURCE_OPTION : string := "auto" -- can be "distributed", "block", or  "auto"
    );
    port (
        enable : in std_logic;
        clock  : in std_logic;
        x_address : out std_logic_vector(X_ADDR_WIDTH-1 downto 0);
        y_address : in std_logic_vector(Y_ADDR_WIDTH-1 downto 0);

        x   : in std_logic_vector(DATA_WIDTH-1 downto 0);
        y  : out std_logic_vector(DATA_WIDTH-1 downto 0);

        done   : out std_logic
    );
end linear;

architecture rtl of linear is
    -----------------------------------------------------------
    -- Functions
    -----------------------------------------------------------
    -- macc
    function multiply_accumulate(w : in signed(DATA_WIDTH-1 downto 0);
                    x : in signed(DATA_WIDTH-1 downto 0);
                    y_0 : in signed(2*DATA_WIDTH-1 downto 0)
            ) return signed is

        variable TEMP : signed(DATA_WIDTH*2-1 downto 0) := (others=>'0');
        variable TEMP2 : signed(DATA_WIDTH-1 downto 0) := (others=>'0');
        variable TEMP3 : signed(FRAC_WIDTH-1 downto 0) := (others=>'0');
    begin
        TEMP := w * x;

        return TEMP+y_0;
    end function;

    function cut_down(x: in signed(2*DATA_WIDTH-1 downto 0))return signed is
        variable TEMP2 : signed(DATA_WIDTH-1 downto 0) := (others=>'0');
        variable TEMP3 : signed(FRAC_WIDTH-1 downto 0) := (others=>'0');
    begin

        TEMP2 := x(DATA_WIDTH+FRAC_WIDTH-1 downto FRAC_WIDTH);
        TEMP3 := x(FRAC_WIDTH-1 downto 0);
        if TEMP2(DATA_WIDTH-1) = '1' and TEMP3 /= 0 then
            TEMP2 := TEMP2 + 1;
        end if;

        if x>0 and TEMP2<0 then
            TEMP2 := ('0', others => '1');
        elsif x<0 and TEMP2>0 then
            TEMP2 := ('1', others => '0');
        end if;
        return TEMP2;
    end function;

    -- Log2 function is for calculating the bitwidth of the address lines
    -- for bias and weights rom
    function log2(val : INTEGER) return natural is
        variable res : natural;
    begin
        for i in 1 to 31 loop
            if (val <= (2 ** i)) then
                res := i;
                exit;
            end if;
        end loop;
        return res;
    end function log2;

    -----------------------------------------------------------
    -- Signals
    -----------------------------------------------------------
    constant FXP_ZERO : signed(DATA_WIDTH-1 downto 0) := (others=>'0');
    constant FXP_ONE : signed(DATA_WIDTH-1 downto 0) := to_signed(2**FRAC_WIDTH,DATA_WIDTH);

    type t_state is (s_stop, s_forward, s_idle);

    signal n_clock : std_logic;
    signal w_in : std_logic_vector(DATA_WIDTH-1 downto 0) := (others=>'0');
    signal b_in : std_logic_vector(DATA_WIDTH-1 downto 0) := (others=>'0');

    signal addr_w : std_logic_vector(log2(IN_FEATURE_NUM*OUT_FEATURE_NUM)-1 downto 0) := (others=>'0');
    --signal addr_b : std_logic_vector((log2(OUT_FEATURE_NUM)-1) downto 0) := (others=>'0');
    signal addr_b : std_logic_vector(Y_ADDR_WIDTH-1 downto 0) := (others=>'0');

    signal fxp_x, fxp_w, fxp_b, fxp_y : signed(DATA_WIDTH-1 downto 0) := (others=>'0');
    signal macc_sum : signed(2*DATA_WIDTH-1 downto 0) := (others=>'0');

    signal reset : std_logic := '0';
    signal state : t_state;

    -- simple solution for the output buffer
    type t_y_array is array (0 to OUT_FEATURE_NUM) of std_logic_vector(DATA_WIDTH-1 downto 0);
    signal y_ram : t_y_array;
    attribute rom_style : string;
    attribute rom_style of y_ram : signal is RESOURCE_OPTION;

begin

    -- connecting signals to ports
    n_clock <= not clock;

    fxp_w <= signed(w_in);
    fxp_x <= signed(x);
    fxp_b <= signed(b_in);

    -- connects ports
    reset <= not enable;

    linear_main : process (clock, enable, reset)
        variable current_neuron_idx : integer range 0 to OUT_FEATURE_NUM-1 := 0;
        variable current_input_idx : integer  range 0 to IN_FEATURE_NUM-1 := 0;
        variable var_addr_w : integer range 0 to OUT_FEATURE_NUM*IN_FEATURE_NUM-1 := 0;
        variable var_sum, var_y : signed(2*DATA_WIDTH-1 downto 0);
        variable var_w, var_x : signed(DATA_WIDTH-1 downto 0);
        variable y_write_en : std_logic;
        variable var_y_write_idx : integer;
    begin

        if (reset = '1') then
            state <= s_stop;
            done <= '0';

            current_neuron_idx := 0;
            current_input_idx := 0;
            var_addr_w := 0;

        elsif rising_edge(clock) then

            if state=s_stop then
                state <= s_forward;

                -- first add b accumulated sum
                var_y := (others=>'0');
                var_x := fxp_b;
                var_w := FXP_ONE;
            elsif state=s_forward then

                -- remapping to x and w
                var_y := macc_sum;
                var_x := fxp_x;
                var_w := fxp_w;

                if current_input_idx<IN_FEATURE_NUM-1 then
                    current_input_idx := current_input_idx + 1;
                    var_addr_w := var_addr_w + 1;
                else
                    current_input_idx := 0;

                    y_write_en := '1';
                    var_y_write_idx := current_neuron_idx;

                    if current_neuron_idx<OUT_FEATURE_NUM-1 then
                        current_neuron_idx := current_neuron_idx + 1;
                        var_addr_w := var_addr_w + 1;
                        state <= s_stop;
                    else
                        state <= s_idle;
                        done <= '1';
                    end if;

                end if;
            end if;

            var_sum := multiply_accumulate(var_w, var_x, var_y);
            macc_sum <= var_sum;

            if y_write_en='1'then
                y_ram(var_y_write_idx) <= std_logic_vector(cut_down(var_sum));
                y_write_en := '0';
            end if;

        end if;

        x_address <= std_logic_vector(to_unsigned(current_input_idx, x_address'length));
        addr_w <= std_logic_vector(to_unsigned(var_addr_w, addr_w'length));
        addr_b <= std_logic_vector(to_unsigned(current_neuron_idx, addr_b'length));
    end process linear_main;

    y_reading : process (clock, state)
    begin
        if (state=s_idle) or (state=s_stop) then
            if falling_edge(clock) then
                -- After the layer in at idle mode, y is readable
                -- but it only update at the rising edge of the clock
                y <= y_ram(to_integer(unsigned(y_address)));
            end if;
        end if;
    end process y_reading;

    -- Weights
    rom_w : entity work.linear_w_rom(rtl)
    port map  (
        clk  => n_clock,
        en   => '1',
        addr => addr_w,
        data => w_in
    );

    -- Bias
    rom_b : entity work.linear_b_rom(rtl)
    port map  (
        clk  => n_clock,
        en   => '1',
        addr => addr_b,
        data => b_in
    );

end architecture rtl;"""

    linear = LinearCreator(
        total_bits=16, frac_bits=8, in_features=3, out_features=2, bias=False
    )

    design = linear.create_design("linear")
    destination = InMemoryPath("linear", parent=None)
    design.save_to(destination)
    actual_linear_code = "\n".join(cast(InMemoryFile, destination["linear"]).text)

    assert expected_linear_code == actual_linear_code


if __name__ == "__main__":
    main()



from elasticai.creator.ir2vhdl import VhdlNode

from .clocked_combinatorial import ClockedInstance
from .node_factory import InstanceFactoryForCombinatorial


@InstanceFactoryForCombinatorial.register
def sliding_window(node):
    return SlidingWindowNode(node)


class SlidingWindowNode(ClockedInstance):
    _logic_signals_with_default_suffix = ("valid_in", "valid_out")

    def __init__(
        self,
        node: VhdlNode,
    ):
        data_width = node.input_shape.depth
        num_points = node.output_shape.depth
        if (
            "generic_map" in node.attributes  #  pyright: ignore
            and "stride" in node.attributes["generic_map"]  #  pyright: ignore
        ):
            stride = node.attributes["generic_map"]["stride"]  # pyright: ignore
        else:
            stride = node.attributes["stride"]  # pyright: ignore
        super().__init__(
            node,
            input_width=data_width,
            output_width=num_points,
            generic_map=dict(
                stride=stride,
                input_width=data_width,
                output_width=num_points,
            ),
        )



from typing import cast

import torch

from elasticai.creator.file_generation.in_memory_path import InMemoryFile, InMemoryPath
from elasticai.creator.nn.fixed_point.precomputed.precomputed_module import (
    PrecomputedModule,
)


def test_vhdl_code_matches_expected_for_tanh_as_base_module() -> None:
    expected = """library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;               -- for type conversions

entity tanh is
    port (
        enable : in std_logic;
        clock  : in std_logic;
        x      : in std_logic_vector(8-1 downto 0);
        y      : out std_logic_vector(8-1 downto 0)
    );
end tanh;

architecture rtl of tanh is
    signal signed_x : signed(8-1 downto 0) := (others=>'0');
    signal signed_y : signed(8-1 downto 0) := (others=>'0');
begin
    signed_x <= signed(x);
    y <= std_logic_vector(signed_y);
    tanh_process : process(x)
    begin
        if signed_x <= -20 then signed_y <= to_signed(-3, 8);
        elsif signed_x <= -10 then signed_y <= to_signed(-3, 8);
        elsif signed_x <= 0 then signed_y <= to_signed(0, 8);
        elsif signed_x <= 10 then signed_y <= to_signed(3, 8);
        else signed_y <= to_signed(3, 8);
        end if;
    end process;
end rtl;
""".splitlines()
    tanh = PrecomputedModule(
        base_module=torch.nn.Tanh(),
        total_bits=8,
        frac_bits=2,
        num_steps=5,
        sampling_intervall=(-5, 5),
    )
    build_path = InMemoryPath("build", parent=None)
    design = tanh.create_design("tanh")
    design.save_to(build_path)
    actual = cast(InMemoryFile, build_path["tanh"]).text
    assert actual == expected


def test_vhdl_code_matches_expected_for_sigmoid_as_base_module() -> None:
    expected = """library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;               -- for type conversions

entity sigmoid is
    port (
        enable : in std_logic;
        clock  : in std_logic;
        x      : in std_logic_vector(8-1 downto 0);
        y      : out std_logic_vector(8-1 downto 0)
    );
end sigmoid;

architecture rtl of sigmoid is
    signal signed_x : signed(8-1 downto 0) := (others=>'0');
    signal signed_y : signed(8-1 downto 0) := (others=>'0');
begin
    signed_x <= signed(x);
    y <= std_logic_vector(signed_y);
    sigmoid_process : process(x)
    begin
        if signed_x <= -20 then signed_y <= to_signed(0, 8);
        elsif signed_x <= -10 then signed_y <= to_signed(0, 8);
        elsif signed_x <= 0 then signed_y <= to_signed(2, 8);
        elsif signed_x <= 10 then signed_y <= to_signed(3, 8);
        else signed_y <= to_signed(3, 8);
        end if;
    end process;
end rtl;
""".splitlines()
    sigmoid = PrecomputedModule(
        base_module=torch.nn.Sigmoid(),
        total_bits=8,
        frac_bits=2,
        num_steps=5,
        sampling_intervall=(-5, 5),
    )
    build_path = InMemoryPath("build", parent=None)
    design = sigmoid.create_design("sigmoid")
    design.save_to(build_path)
    actual = cast(InMemoryFile, build_path["sigmoid"]).text
    assert actual == expected



from .base_graph import BaseGraph
from .graph import Graph
from .graph_iterators import (
    bfs_iter_down,
    bfs_iter_up,
    dfs_pre_order,
)
from .graph_rewriting import GraphRewriter, RewriteResult
from .name_generation import NameRegistry
from .subgraph_matching import find_subgraphs

__all__ = [
    "BaseGraph",
    "Graph",
    "find_subgraphs",
    "GraphRewriter",
    "bfs_iter_down",
    "bfs_iter_up",
    "dfs_pre_order",
    "NameRegistry",
    "RewriteResult",
]



from typing import Callable

import pytest
import torch

from elasticai.creator.nn.fixed_point.linear.design import LinearDesign
from elasticai.creator.nn.fixed_point.linear.testbench import LinearTestbench
from elasticai.creator.nn.fixed_point.number_converter import FXPParams, NumberConverter
from elasticai.creator.vhdl.auto_wire_protocols.port_definitions import create_port
from elasticai.creator.vhdl.design.ports import Port


class DummyLinear:
    def __init__(
        self, fxp_params: FXPParams, in_signal_length: int, out_signal_length: int
    ):
        self.name: str = "linear0"
        self.in_feature_num = in_signal_length
        self.out_feature_num = out_signal_length
        self.data_width = fxp_params.total_bits
        self.frac_width = fxp_params.frac_bits
        self.port: Port = create_port(
            y_width=fxp_params.total_bits,
            x_width=fxp_params.total_bits,
            x_count=1,
            y_count=2,
        )


def parameters_for_reported_content_parsing(fxp_params, input_expected_pairs):
    def add_expected_prefix_to_pairs(pairs):
        _converter_for_batch = NumberConverter(
            FXPParams(8, 0)
        )  # max for 255 lines of inputs
        pairs_with_prefix = list()
        for i, (pairs_text, pairs_number) in enumerate(pairs):
            pairs_with_prefix.append(list())
            pairs_with_prefix[i].append(list())
            pairs_with_prefix[i].append(pairs_number)
            for batch_number, batch_channel_text in enumerate(pairs_text):
                for out_channel_text in batch_channel_text:
                    for value_text in out_channel_text:
                        pairs_with_prefix[i][0].append(
                            f"result: {_converter_for_batch.integer_to_bits(batch_number)},"
                            f" {value_text}"
                        )
        return pairs_with_prefix

    pairs_with_prefix = [
        (fxp_params, a, b)
        for a, b in add_expected_prefix_to_pairs(input_expected_pairs)
    ]
    return pairs_with_prefix


@pytest.fixture
def create_uut() -> Callable[[FXPParams, int, int], LinearDesign]:
    def create(
        fxp_params: FXPParams, in_signal_length: int, out_signal_length: int
    ) -> LinearDesign:
        return DummyLinear(
            fxp_params=fxp_params,
            in_signal_length=in_signal_length,
            out_signal_length=out_signal_length,
        )

    return create


@pytest.mark.parametrize(
    "fxp_params, reported, y",
    (
        parameters_for_reported_content_parsing(
            fxp_params=FXPParams(total_bits=3, frac_bits=0),
            input_expected_pairs=[
                ([[["010"]]], [[[2.0]]]),
                ([[["001", "010"]]], [[[1.0, 2.0]]]),
                ([[["111", "001"]]], [[[-1.0, 1.0]]]),
            ],
        )
        + parameters_for_reported_content_parsing(
            fxp_params=FXPParams(total_bits=4, frac_bits=1),
            input_expected_pairs=[
                ([[["0001", "1111"]]], [[[0.5, -0.5]]]),
                ([[["0001", "0011", "1000", "1111"]]], [[[0.5, 1.5, -4.0, -0.5]]]),
            ],
        )
    ),
)
def test_parse_reported_content_one_out_channel(fxp_params, reported, y, create_uut):
    in_signal_length = None
    out_signal_length = 1
    bench = LinearTestbench(
        name="linear_testbench",
        uut=create_uut(fxp_params, in_signal_length, out_signal_length),
    )
    print(f"{reported=}")
    assert y == bench.parse_reported_content(reported)


def test_input_preparation_with_one_in_channel(create_uut):
    fxp_params = FXPParams(total_bits=3, frac_bits=0)
    in_signal_length = 1
    out_signal_length = None
    bench = LinearTestbench(
        name="linear_testbench",
        uut=create_uut(fxp_params, in_signal_length, out_signal_length),
    )
    input = torch.Tensor([[[1.0, 1.0]]])
    expected = [{"x_0_0": "001", "x_0_1": "001"}]
    assert expected == bench.prepare_inputs(input.tolist())



from elasticai.creator.nn.fixed_point.number_conversion import (
    bits_to_rational,
    convert_rational_to_bit_pattern,
    integer_to_bits,
    max_rational,
    min_rational,
)


def test_001_to_0_5():
    pattern = "001"
    assert 0.5 == bits_to_rational(pattern, frac_bits=1)


def test_bits_to_rational_strips_whitespace():
    pattern = " 100"
    assert -4 == bits_to_rational(pattern, frac_bits=0)


def test_001_to_0_25():
    pattern = "001"
    assert 0.25 == bits_to_rational(pattern, frac_bits=2)


def test_100_to_minus_4():
    assert -4 == bits_to_rational("100", frac_bits=0)


def test_100_to_minus_2():
    assert -2 == bits_to_rational("100", frac_bits=1)


def test_110_to_minus_1():
    pattern = "110"
    assert -1 == bits_to_rational(pattern, frac_bits=1)


def test_1_to_minus_1():
    pattern = "1"
    assert -1 == bits_to_rational(pattern, frac_bits=0)


def test_1100_to_minus_1():
    pattern = "1100"
    assert -1 == bits_to_rational(pattern, frac_bits=2)


def test_10_to_minus_2():
    pattern = "10"
    assert -2 == bits_to_rational(pattern, frac_bits=0)


def test_01_to_1():
    pattern = "01"
    assert 1 == bits_to_rational(pattern, frac_bits=0)


def test_1_to_01():
    assert "01" == integer_to_bits(number=1, total_bits=2)


def test_2_to_10():
    assert "10" == integer_to_bits(number=2, total_bits=2)


def test_minus_5_to_1011():
    assert "1011" == integer_to_bits(number=-5, total_bits=4)


def test_minus_8_to_1000():
    assert "1000" == integer_to_bits(number=-8, total_bits=4)


def test_0_5_to_01():
    assert "01" == convert_rational_to_bit_pattern(
        rational=0.5, total_bits=2, frac_bits=1
    )


def test_0_5_to_001():
    assert "001" == convert_rational_to_bit_pattern(
        rational=0.5, total_bits=3, frac_bits=1
    )


def test_0_25_to_000_because_of_rounding():
    assert "000" == convert_rational_to_bit_pattern(
        rational=0.25, total_bits=3, frac_bits=1
    )


def test_max_rational_for_3_and_1_is_1_5():
    assert 1.5 == max_rational(total_bits=3, frac_bits=1)


def test_max_rational_for_3_and_2_is_0_75():
    assert 0.75 == max_rational(3, 2)


def test_min_rational_for_3_and_2_is_minus_1_75():
    assert -1.0 == min_rational(3, 2)


def test_min_rational_for_3_and_1_is_minus_3_5():
    assert -2 == min_rational(3, 1)



import copy
from collections.abc import Mapping, Sequence
from typing import Protocol

from .graph import Graph
from .name_generation import NameRegistry


class RewriteResult:
    def __init__(
        self,
        *,
        new_graph: Graph[str],
        pattern_to_original: dict[str, str],
        replacement_to_new: dict[str, str],
    ) -> None:
        self.new_graph = new_graph
        self.pattern_to_original = pattern_to_original
        self.replacement_to_new = replacement_to_new


class Matcher(Protocol):
    def __call__(self, *, pattern: Graph, graph: Graph) -> Sequence[dict[str, str]]: ...


class GraphRewriter:
    """Rewrite graphs based on a pattern, interface, and replacement.

    The terminology here is based on the double pushout approach to graph rewriting.
    The algorithm will find the first occurence of pattern in the graph using the `match` function.
    Then, it will create a new graph by replacing the pattern with the replacement.
    The structures specified by `interface` are excluded from this replacement.
    Instead the interface is used as the "glue" between the original graph and the replacement.
    The functions `lhs` and `rhs` are used to identify the `interface` nodes in the `pattern` and `replacement` respectively.

    :::{important}
    * The `rhs` function is required to be injective. This means that each interface node should be mapped to a unique replacement node.
    * The algorithm will stop on the first match. If you need to replace multiple matches you have to call the `rewrite` function multiple times.
    * The nodes from `replacement` will be automatically renamed to avoid conflicts with the nodes in the original graph, i.e., if node `a` exists already we will add a new node `a_1`.
    :::

    :param pattern: The pattern to match in the graph.
    :param interface: The interface where graph and replacement will be "glued" together.
    :param replacement: The replacement for the pattern.
    :param match: A function that takes a graph and a pattern and returns a dictionary of matches. See [`find_subgraphs`](#elasticai.creator.ir.find_subgraphs) for more information.
    :param lhs: A dict that is used to map the interface to the pattern.
    :param rhs: A dict that is used to map the interface to the replacement.


    """

    def __init__(
        self,
        *,
        pattern: Graph[str],
        interface: Graph[str],
        replacement: Graph[str],
        match: Matcher,
        lhs: Mapping[str, str],
        rhs: Mapping[str, str],
    ) -> None:
        self._pattern = pattern
        self._interface = interface
        self._replacement = replacement
        self._match = match
        self._lhs = lhs
        self._rhs = rhs
        self._rhs_inversed = {self._rhs[node]: node for node in self._interface.nodes}
        self._replacement_nodes_in_interface = self._rhs_inversed.keys()
        if len(self._interface.nodes) != len(self._replacement_nodes_in_interface):
            raise ValueError(
                "Ensure the `rhs` function is injective. The `rhs` function should map each interface node to a unique replacement node."
            )

    def rewrite(self, graph: Graph[str]) -> RewriteResult:
        """Rewrite the graph based on the pattern, interface, and replacement.

        Returns a [`RewriteResult`](#elasticai.creator.ir.graph_rewriting.RewriteResult) object containing the new graph and two dicts mapping nodes from `pattern` to the original graph and nodes from `replacement` to the new graph.
        matches = self._match(graph, self._pattern)
        """
        matches = self._match(pattern=self._pattern, graph=graph)
        if len(matches) > 0:
            match = matches[0]
        else:
            return RewriteResult(
                new_graph=copy.deepcopy(graph),
                pattern_to_original={},
                replacement_to_new={},
            )
        interface_nodes_in_pattern = set(
            self._lhs[node] for node in self._interface.nodes
        )
        interface_nodes_in_graph = set(
            match[node] for node in interface_nodes_in_pattern
        )

        new_graph: Graph = graph.new()
        nodes_to_be_removed = set(match.values()) - set(interface_nodes_in_graph)
        nodes_to_keep = set(graph.nodes) - nodes_to_be_removed
        name_registry = SingleNewNameGenerator(
            NameRegistry().prepopulate(nodes_to_keep)
        )

        def get_replacement_node_in_new_graph(node: str) -> str:
            if node in self._replacement_nodes_in_interface:
                return match[self._lhs[self._rhs_inversed[node]]]
            if node in nodes_to_keep:
                return name_registry.get_name(node)
            return node

        def get_replacement_edge_for_new_graph(src: str, sink: str) -> tuple[str, str]:
            src = get_replacement_node_in_new_graph(src)
            sink = get_replacement_node_in_new_graph(sink)
            return src, sink

        for node in nodes_to_keep:
            new_graph.add_node(node)

        for node in self._replacement.nodes:
            if node not in self._replacement_nodes_in_interface:
                new_graph.add_node(get_replacement_node_in_new_graph(node))

        for src, sink in graph.iter_edges():
            if src in nodes_to_keep and sink in nodes_to_keep:
                new_graph.add_edge(src, sink)

        for src, sink in self._replacement.iter_edges():
            new_graph.add_edge(*get_replacement_edge_for_new_graph(src, sink))

        return RewriteResult(
            new_graph=new_graph,
            pattern_to_original=match,
            replacement_to_new={
                r: get_replacement_node_in_new_graph(r) for r in self._replacement.nodes
            },
        )


class SingleNewNameGenerator:
    def __init__(self, registry: NameRegistry) -> None:
        self._registry = registry
        self._reversed: dict[str, str] = {}
        self.memory: dict[str, str] = {}

    def get_name(self, name: str) -> str:
        if name in self.memory:
            return self.memory[name]
        new_name = self._registry.get_unique_name(name)
        self.memory[name] = new_name
        self._reversed[new_name] = name
        return new_name



from typing import cast

import pytest
import torch

from elasticai.creator.nn.fixed_point.round_to_fixed_point import RoundToFixedPoint
from elasticai.creator.nn.fixed_point.two_complement_fixed_point_config import (
    FixedPointConfig,
)
from tests.tensor_test_case import assertTensorEqual


def roundToFxp(inputs: list[float] | torch.Tensor) -> torch.Tensor:
    if isinstance(inputs, list):
        inputs = torch.tensor(inputs)
    config = FixedPointConfig(total_bits=4, frac_bits=2)
    return cast(torch.Tensor, RoundToFixedPoint.apply(inputs, config))


def test_round_upper_bound() -> None:
    assertTensorEqual(
        expected=roundToFxp([1.75]),
        actual=[1.75],
    )


def test_round_lower_bound() -> None:
    assertTensorEqual(
        expected=roundToFxp([-2.0]),
        actual=[-2.0],
    )


def test_round_out_of_bounds() -> None:
    inputs = torch.tensor([2.0])
    with pytest.raises(ValueError):
        _ = roundToFxp(inputs)


def test_round_typical_values() -> None:
    assertTensorEqual(
        expected=roundToFxp([-1.3, 0.1, 1.6]),
        actual=[-1.25, 0, 1.5],
    )



import pytest

from elasticai.creator.graph import BaseGraph as _Graph
from elasticai.creator.ir import Attribute, Edge, Implementation, Node, edge, node


class Graph(Implementation[Node, Edge]):
    def __init__(self, data: dict[str, Attribute] | None = None):
        g: _Graph[str] = _Graph()
        if data is None:
            data = {}
        super().__init__(
            graph=g,
            node_fn=Node,
            edge_fn=Edge,
            data=data,
        )


def test_graph_is_serialized():
    g = Graph(data=dict(name="network", type="network"))
    g.add_node(node(name="node1", type="type1"))
    g.add_node(node(name="node2", type="type2"))
    g.add_edge(edge(src="node1", sink="node2"))
    assert g.data == {
        "name": "network",
        "type": "network",
        "nodes": {
            "node1": {"name": "node1", "type": "type1"},
            "node2": {"name": "node2", "type": "type2"},
        },
        "edges": {
            ("node1", "node2"): {"src": "node1", "sink": "node2"},
        },
    }


def test_graph_has_required_fields():
    g = Graph(data=dict(name="network", type="network"))
    assert g.name == "network"


def test_can_retrieve_edges() -> None:
    g = (
        Graph()
        .add_node(name="x", type="t")
        .add_node(name="y", type="t")
        .add_edge(src="x", sink="y")
    )
    edges = tuple(g.edges.values())
    assert edge(src="x", sink="y") == edges[0]


@pytest.fixture
def graph() -> Graph:
    g = (
        Graph()
        .add_node(name="x", type="t")
        .add_node(name="y", type="t")
        .add_node(name="z", type="t")
        .add_edge(src="x", sink="y")
        .add_edge(src="y", sink="z", extra="e")
    )
    return g


def test_can_get_specific_edge(graph: Graph) -> None:
    e = graph.edges[("y", "z")]
    assert dict(extra="e") == e.attributes


def test_edges_is_read_only(graph: Graph) -> None:
    with pytest.raises(TypeError):
        graph.edges[("x", "z")] = edge("x", "z")  # type: ignore


def test_predecessors_of_z_is_y(graph: Graph) -> None:
    n = next(iter(graph.predecessors("z").values()))
    assert graph.nodes["y"] == n


def test_successor_of_x_is_y(graph: Graph) -> None:
    n = next(iter(graph.successors("x").values()))
    assert graph.nodes["y"] == n



import pytest

from elasticai.creator.graph.graph_rewriting import RewriteResult
from tests.unit_tests.graph.utils import Graph

from .utils import (
    build_graph_from_dict,
    get_rewriter,
    get_rewriter_returning_full_result,
)


def test_can_replace_single_node():
    graph = build_graph_from_dict(
        {("a", "a_t"): ["b"], ("b", "b_t"): ["c"], ("c", "c_t"): []}
    )
    pattern = build_graph_from_dict(
        {("0", "a_t"): ["1"], ("1", "b_t"): ["2"], ("2", "c_t"): []}
    )
    interface = build_graph_from_dict({("i0", ""): [], ("i1", ""): []})
    replacement = build_graph_from_dict(
        {("r0", "a_t"): ["r1"], ("r1", "r_t"): ["r2"], ("r2", "c_t"): []}
    )

    lhs = {"i0": "0", "i1": "2"}

    rhs = {"i0": "r0", "i1": "r2"}

    rewrite = get_rewriter(pattern, interface, replacement, lhs, rhs)

    expected = build_graph_from_dict(
        {
            ("c", "c_t"): [],
            ("a", "a_t"): ["r1"],
            ("r1", "r_t"): ["c"],
        }
    )
    assert rewrite(graph).as_dict() == expected.as_dict()


def test_create_new_name_for_replacement_in_case_of_conflict():
    graph = build_graph_from_dict(
        {("a", "a_t"): ["b"], ("b", "b_t"): ["c"], ("c", "c_t"): []}
    )
    pattern = build_graph_from_dict(
        {("0", "a_t"): ["1"], ("1", "b_t"): ["2"], ("2", "c_t"): []}
    )
    interface = build_graph_from_dict({("i0", ""): [], ("i1", ""): []})
    replacement = build_graph_from_dict(
        {("r0", "a_t"): ["a"], ("a", "r_t"): ["r2"], ("r2", "c_t"): []}
    )

    lhs = {"i0": "0", "i1": "2"}

    rhs = {"i0": "r0", "i1": "r2"}

    rewrite = get_rewriter(pattern, interface, replacement, lhs, rhs)

    expected = build_graph_from_dict(
        {
            ("c", "c_t"): [],
            ("a", "a_t"): ["a_1"],
            ("a_1", "r_t"): ["c"],
        }
    )
    assert rewrite(graph).as_dict() == expected.as_dict()


def test_raise_error_if_rhs_is_not_injective():
    pattern = build_graph_from_dict(
        {("0", "a_t"): ["1"], ("1", "b_t"): ["2"], ("2", "c_t"): []}
    )
    interface = build_graph_from_dict({("i0", ""): [], ("i1", ""): []})
    replacement = build_graph_from_dict(
        {("r0", "a_t"): ["a"], ("a", "r_t"): ["r2"], ("r2", "c_t"): []}
    )

    lhs = {"i0": "0", "i1": "2"}
    rhs = {"i0": "r0", "i1": "r0"}

    pytest.raises(
        ValueError,
        get_rewriter,
        pattern,
        interface,
        replacement,
        lhs,
        rhs,
    )


def test_can_replace_one_of_two_matches():
    graph = build_graph_from_dict(
        {
            ("a", "a_t"): ["b"],
            ("b", "b_t"): ["c"],
            ("c", "a_t"): ["d"],
            ("d", "b_t"): ["e"],
            ("e", "a_t"): [],
        }
    )
    pattern = build_graph_from_dict(
        {("0", "a_t"): ["1"], ("1", "b_t"): ["2"], ("2", "a_t"): []}
    )
    interface = build_graph_from_dict({("i0", ""): [], ("i1", ""): []})
    replacement = build_graph_from_dict(
        {("r0", "a_t"): ["r1"], ("r1", "r_t"): ["r2"], ("r2", "c_t"): []}
    )

    lhs = {"i0": "0", "i1": "2"}
    rhs = {"i0": "r0", "i1": "r2"}

    rewrite = get_rewriter(pattern, interface, replacement, lhs, rhs)

    expected = build_graph_from_dict(
        {
            ("c", "a_t"): ["r1"],
            ("a", "a_t"): ["b"],
            ("b", "b_t"): ["c"],
            ("r1", "r_t"): ["e"],
            ("e", "a_t"): [],
        }
    )
    actual = rewrite(graph)
    assert set(actual.edges) == set(expected.edges)
    assert set(actual.nodes) == set(expected.nodes)


class TestRewriteResult:
    @pytest.fixture
    def graph(self):
        return build_graph_from_dict(
            {("a", "a_t"): ["b"], ("b", "b_t"): ["c"], ("c", "c_t"): []}
        )

    @pytest.fixture
    def pattern(self):
        return build_graph_from_dict(
            {("0", "a_t"): ["1"], ("1", "b_t"): ["2"], ("2", "c_t"): []}
        )

    @pytest.fixture
    def replacement(self):
        return build_graph_from_dict(
            {("r0", "a_t"): ["r1"], ("r1", "r_t"): ["r2"], ("r2", "c_t"): []}
        )

    @pytest.fixture
    def rhs_dict(self):
        return {"i0": "r0", "i1": "r2"}

    @pytest.fixture
    def result(
        self,
        graph: Graph,
        pattern: Graph,
        replacement: Graph,
        rhs_dict: dict[str, str],
    ):
        interface = build_graph_from_dict({("i0", ""): [], ("i1", ""): []})

        lhs = {"i0": "0", "i1": "2"}

        rhs = rhs_dict

        rewrite = get_rewriter_returning_full_result(
            pattern, interface, replacement, lhs, rhs
        )
        result = rewrite(graph)
        return result

    def test_can_map_pattern_back_to_original_graph(self, result: RewriteResult):
        expected_pattern_to_original = {"0": "a", "1": "b", "2": "c"}

        assert result.pattern_to_original == expected_pattern_to_original

    def test_can_map_replacement_to_new_graph(self, result: RewriteResult):
        expected_replacement_to_new = {"r0": "a", "r1": "r1", "r2": "c"}

        assert result.replacement_to_new == expected_replacement_to_new

    def test_can_combine_attribute_from_replacement_and_original_graph(
        self,
        result: RewriteResult,
        graph: Graph,
        pattern: Graph,
        replacement: Graph,
        rhs_dict: dict[str, str],
    ):
        graph.data = {"a": "1", "b": "2", "c": "3"}
        replacement.data = {"r0": "100", "r1": "200", "r2": "300"}

        new_graph = Graph(result.new_graph, {})

        for repl_node in replacement.data:
            new_node = result.replacement_to_new[repl_node]
            original_node_with_interesting_value = result.pattern_to_original["1"]
            interesting_value = graph.data[original_node_with_interesting_value]
            replacement_value = replacement.data[repl_node]
            not_in_interface = repl_node not in rhs_dict.values()
            if not_in_interface:
                new_graph.data[new_node] = str(
                    int(replacement_value) + int(interesting_value)
                )
            else:
                new_graph.data[new_node] = graph.data[new_node]

        assert new_graph.data == {"a": "1", "r1": "202", "c": "3"}



from string import Template as _pyTemplate
from typing import Mapping, Self, cast

from elasticai.creator import template as tpl


class _Definition(tpl.TemplateParameter):
    def __init__(self, type: str, name: str, delimiter: str) -> None:
        self.name = name
        self.delimiter = delimiter
        self.regex = r"(?P<param>{type}\s+(signed\s+)?(\[.*?\]\s+)?{name}\s*=)\s?.[^;,\n]*".format(
            type=type, name=name
        )

    def replace(self, m: dict[str, str]) -> str:
        d = self.delimiter
        return f"{m['param']} {d}{self.name}"


class _LocalParameter(_Definition):
    """Turn the localparam `name` into a template parameter."""

    def __init__(self, name: str, delimiter: str):
        super().__init__("localparam", name, delimiter)


class _Parameter(_Definition):
    def __init__(self, name: str, delimiter: str):
        super().__init__("parameter", name, delimiter)


class _IdParameter(tpl.TemplateParameter):
    def __init__(self, name: str, delimiter: str):
        self.name = name
        self.delimiter = delimiter
        self.regex = r"(?P<prefix>[^_a-zA-Z0-9]*)(?P<id>{name})(?P<suffix>[^_a-zA-Z0-9]*)".format(
            name=name
        )

    def replace(self, m: dict[str, str]) -> str:
        d = self.delimiter
        return f"{m['prefix']}{d}{self.name}{m['suffix']}"


class _DefineSwitch(tpl.TemplateParameter):
    def __init__(self, name: str, delimiter: str):
        self.name = name
        self.regex = r"`define\s+{name}()(?=\s)".format(name=name)
        self.delimiter = delimiter

    def replace(self, m: dict[str, str]) -> str:
        d = self.delimiter
        return f"{d}def{self.name}"


class _ModuleName(tpl.AnalysingTemplateParameter):
    def __init__(self, delimiter: str):
        self._pre_analysis_regex = (
            r"(?P<prefix>[^_a-zA-Z0-9]){name}(?P<suffix>[^_a-zA-Z0-9])"
        )
        self.regex = ""
        self.analyse_regex = r"module\s+(?P<name>[a-zA-Z_][a-zA-Z0-9_]*)"
        self._analysed = False
        self._delimiter = delimiter

    def analyse(self, match: dict[str, str]) -> None:
        self._analysed = True
        name = match["name"]
        self.regex = self._pre_analysis_regex.format(name=name)

    def replace(self, match: dict[str, str]) -> str:
        d = self._delimiter
        if not self._analysed:
            raise ValueError("No module name found")
        return f"{match['prefix']}{d}module_name{match['suffix']}"


class TemplateDirector:
    """Director for verilog templates.

    Most methods correspond to verilog language constructs."""

    def __init__(self) -> None:
        self._builder = tpl.TemplateBuilder()
        self._define_switches: dict[str, bool] = {}
        self._delimiter = "§"

    def reset(self) -> Self:
        self._builder = tpl.TemplateBuilder()
        self._define_switches = {}
        return self

    def set_prototype(self, prototype: str) -> Self:
        self._builder.set_prototype(prototype)
        return self

    def parameter(self, name: str) -> Self:
        self._builder.add_parameter(_Parameter(name, self._delimiter))
        return self

    def localparam(self, name: str) -> Self:
        self._builder.add_parameter(_LocalParameter(name, self._delimiter))
        return self

    def define_scoped_switch(self, name: str, default=True) -> Self:
        """Add a switch for a define that is scoped to the module name.

        The switch will be prefixed with the value that users provide as `module_name` to the render call.
        """
        switch = _DefineSwitch(name, self._delimiter)
        self._builder.add_parameter(switch)
        self._builder.add_parameter(_IdParameter(name, self._delimiter))
        self._define_switches[name] = default
        return self

    def add_module_name(self) -> Self:
        self._builder.add_parameter(_ModuleName(self._delimiter))
        return self

    def build(self) -> "VerilogTemplate":
        MyStrTemplate = type(
            "MyStrTemplate", (_pyTemplate,), {"delimiter": self._delimiter}
        )
        return VerilogTemplate(
            MyStrTemplate(self._builder.build()), self._define_switches
        )


class VerilogTemplate:
    def __init__(self, template: _pyTemplate, defines: dict[str, bool]) -> None:
        self._template = template
        self._defines = defines.copy()

    def undef(self, name: str):
        """Disable a define switch, ie. comment out the define statement"""
        self._defines[name] = False

    def define(self, name: str):
        """Enable a define switch, ie. uncomment the define statement"""
        self._defines[name] = True

    def substitute(self, params: Mapping[str, str | bool]) -> str:
        new_params: dict[str, str] = {}
        if "module_name" in params:
            module_prefix = f"{params['module_name']}_"
        else:
            module_prefix = ""

        for name, value in params.items():
            if name not in self._defines:
                new_params[name] = cast(str, value)

        defines: dict[str, bool] = self._defines.copy()
        for name, enabled in params.items():
            if name in self._defines and isinstance(enabled, bool):
                defines[name] = enabled

        for name, enabled in defines.items():
            defname = f"def{name}"
            new_params[defname] = f"`define {module_prefix}{name}"
            new_params[name] = f"{module_prefix}{name}"
            if not enabled:
                new_params[defname] = "// automatically disabled\n// " + cast(
                    str, new_params[defname]
                )
        return self._template.substitute(new_params)
